import json
import os

import numpy as np

from beartype.typing import Dict, List, Union
from datasets import load_dataset
from chatllama.rlhf.config import Config, ConfigActor, ConfigReward
from chatllama.rlhf.reward import RewardModel, CriticModel
from chatllama.rlhf.actor import ActorModel


ConfigType = Union[Config, ConfigActor, ConfigReward]


class BaseDataset:
    def __init__(
        self,
    ) -> None:
        pass

    @staticmethod
    def sort_conversation(
        conversations: List[Dict],
        only_input: bool = False,
        reverse: bool = True,
        shuffle: bool = True,
    ) -> List[Dict]:
        """Sort the conversations by length of user_input + completion
        or by length of user_input only

        Args:
            conversations (List[Dict]): list of conversations
            only_input (bool, optional): sort by length of user_input only.
                Defaults to False.
            reverse (bool, optional): sort in descending order.
                Defaults to True.
            shuffle (bool, optional): shuffle the dataset leaving only the
                first 100 samples sorted. Defaults to True.

        Returns:
            List[Dict]: sorted list of conversations
        """

        # define the sorting function
        if only_input is True:

            def sort_fun(x):
                return len(x["user_input"])

        else:

            def sort_fun(x):
                return len(x["user_input"]) + len(x["completion"])

        # sort
        conversations = sorted(
            conversations,
            key=sort_fun,
            reverse=reverse,
        )

        # shuffle
        if shuffle is True:
            conversations = (
                conversations[:10]
                + np.random.choice(
                    conversations[10:],
                    size=len(conversations[10:]),
                    replace=False,
                ).tolist()
            )

        return conversations

    @staticmethod
    def take_n_samples(
        conversations: List[Dict],
        n: int,
    ) -> List[Dict]:
        """Take N samples from the dataset

        Args:
            conversations (List[Dict]): list of conversations
            n (int): number of samples to take randomly

        Returns:
            List[Dict]: list of N samples
        """

        # sample N number of index from 0 to len(conversations)
        indexes = np.random.choice(len(conversations), size=n, replace=False)
        # take the samples
        conversations = [conversations[i] for i in indexes]
        return conversations

    @staticmethod
    def clean_dataset(config: ConfigType):
        """Clean the datasets by removing too long examples
        The Reward Dataset constraints are:
        - user_input + completion < Reward model max sequence length
        The Actor Dataset constraints are:
        - user_input + completion < Actor model max sequence length
        The RLHF Training Dataset constraints are:
        - user_input + min_completion < Actor model max sequence length
        - user_input + min_completion < Critic model max sequence length
        - user_input + min_completion < Reward model max sequence length

        Args:
            config (Config): config object
        """

        if isinstance(config, Config):
            print("Start cleaning the dataset for RLHF")
            # constraints
            r_model_max_seq_len = config.reward.max_sequence_length
            a_model_max_seq_len = config.actor.max_sequence_length
            c_model_max_seq_len = config.critic.max_sequence_length
            min_completion = config.actor.min_tokens
            # dataset
            dataset_path = config.trainer.examples_path
            # tokenizers
            r_tokenizer = RewardModel.load_tokenizer(config.reward)
            a_tokenizer = ActorModel.load_tokenizer(config.actor)
            c_tokenizer = CriticModel.load_tokenizer(config.critic)
            # safety tokens
            safety_tokens = config.actor.additonal_prompt_tokens

        elif isinstance(config, ConfigActor):
            print("Start cleaning the dataset for Actor")
            # constraint
            a_model_max_seq_len = config.max_sequence_length
            # dataset
            dataset_path = config.train_dataset_path
            # tokenizer
            a_tokenizer = ActorModel.load_tokenizer(config)
            # safety tokens
            safety_tokens = config.additonal_prompt_tokens

        elif isinstance(config, ConfigReward):
            print("Start cleaning the dataset for Reward")
            # constraint
            r_model_max_seq_len = config.max_sequence_length
            # dataset
            dataset_path = config.train_dataset_path
            # tokenizer
            r_tokenizer = RewardModel.load_tokenizer(config)

        # if there is the datasets
        if os.path.exists(dataset_path):

            # load the dataset
            with open(dataset_path, "r") as f:
                conversations = json.load(f)

            # sort in desceding order - longest first
            if isinstance(config, Config):
                conversations = BaseDataset.sort_conversation(
                    conversations,
                    only_input=True,
                    reverse=True,
                )
            else:
                conversations = BaseDataset.sort_conversation(
                    conversations,
                    only_input=False,
                    reverse=True,
                )

            old_len = len(conversations)
            # remove too long examples
            # since datasets are ordered by the length
            # we can remove the first elements until we find
            # an example that is not too long
            while len(conversations) > 0:

                # get the text to be tokenized
                if isinstance(config, Config):
                    text = conversations[0]["user_input"]
                else:
                    text = (
                        conversations[0]["user_input"]
                        + conversations[0]["completion"]
                    )

                # remove elements from RLHF dataset
                if isinstance(config, Config):
                    a_tokens = a_tokenizer.encode(text, truncation=False)
                    r_tokens = r_tokenizer.encode(text, truncation=False)
                    c_tokens = c_tokenizer.encode(text, truncation=False)
                    if (
                        len(a_tokens) + min_completion + safety_tokens
                        > a_model_max_seq_len
                    ):
                        conversations.pop(0)
                    elif (
                        len(r_tokens) + min_completion + safety_tokens
                        > r_model_max_seq_len
                    ):
                        conversations.pop(0)
                    elif (
                        len(c_tokens) + min_completion + safety_tokens
                        > c_model_max_seq_len
                    ):
                        conversations.pop(0)
                    else:
                        break

                # remove elements from Actor dataset
                elif isinstance(config, ConfigActor):
                    tokens = a_tokenizer.encode(text, truncation=False)
                    if len(tokens) + safety_tokens > a_model_max_seq_len:
                        conversations.pop(0)
                    else:
                        break

                # remove elements from Reward dataset
                elif isinstance(config, ConfigReward):
                    tokens = r_tokenizer.encode(text, truncation=False)
                    if len(tokens) > r_model_max_seq_len:
                        conversations.pop(0)
                    else:
                        break

            # if the number of examples has changed
            if len(conversations) != old_len:
                print("Number of examples before cleaning: ", old_len)
                print(
                    "Number of examples after cleaning: ", len(conversations)
                )

                # remove the old dataset
                os.remove(dataset_path)

                # save the new dataset
                with open(dataset_path, "w") as f:
                    json.dump(conversations, f, indent=4)
            else:
                print("Dataset is already clean")

        else:
            print(
                f"Dataset not found at {dataset_path}"
                f" Skipping cleaning of the dataset"
            )


class StanfordNLPSHPDataset(BaseDataset):
    """Class for Stanford NLP SHP dataset from HuggingFace"""

    def __init__(
        self,
    ) -> None:
        print("Download the dataset")
        self.dataset = load_dataset("stanfordnlp/SHP")
        print("Download Completed")

    def reformat_dataset(self, data: List) -> List[Dict]:
        """Reformat the dataset to the format required by RLHF

        Args:
            data (List): dataset from HuggingFace

        Returns:
            List[Dict]: reformatted dataset
        """

        # initialize conversations
        conversations = []

        # loop over the dataset
        for i, d in enumerate(data):
            if d["score_A"] > d["score_B"]:
                response = d["human_ref_A"]
            else:
                response = d["human_ref_B"]

            # compose user_input template
            user_input = d["history"].rstrip("\n")
            user_input = "Human: " + d["history"] + "\n\n##\n\n"

            # compose completion template
            completion = "Assistant: " + response
            conv = {
                "user_input": user_input,
                "completion": completion,
                "score": None,
            }
            conversations.append(conv)

        return conversations

    def save_dataset(
        self, dataset_folder: str, number_of_samples: int, reverse: bool = True
    ) -> None:
        """Save the dataset in the format required by RLHF

        Args:
            dataset_folder (str): path to the folder where the dataset
                will be saved
            number_of_samples (int): number of samples to take from the
                dataset
            reverse (bool, optional): sort the dataset in descending order.
                Defaults to True.
        """

        print("Generate datasets for RLHF")

        # take the train and test dataset to create the finetuning dataset
        conversations = self.reformat_dataset(self.dataset["train"])
        conversations.extend(self.reformat_dataset(self.dataset["test"]))

        # sort conversations by length of user_input + completion
        conversations = self.sort_conversation(conversations, reverse=reverse)

        # save actor training data
        with open(f"{dataset_folder}/actor_training_data.json", "w") as f:
            json.dump(conversations, f, indent=4)

        # take N samples and sort them
        conversations = self.take_n_samples(conversations, number_of_samples)
        conversations = self.sort_conversation(conversations, reverse=reverse)

        # save reward training data
        with open(f"{dataset_folder}/reward_training_data.json", "w") as f:
            json.dump(conversations, f, indent=4)

        # take the validation dataset for rlhf
        conversations = self.reformat_dataset(self.dataset["validation"])
        # sort the validation dataset
        conversations = self.sort_conversation(
            conversations,
            only_input=True,
            reverse=reverse,
        )
        # save rlhf training data
        with open(f"{dataset_folder}/rlhf_training_data.json", "w") as f:
            json.dump(conversations, f, indent=4)

        print("Generation Completed")


class AnthropicRLHF(BaseDataset):
    def __init__(
        self,
    ) -> None:

        print("Download the dataset")
        self.dataset = load_dataset("Anthropic/hh-rlhf")
        print("Download Completed")

    def reformat_dataset(self, data: List) -> List[Dict]:
        """Reformat the dataset to the format required by RLHF

        Args:
            data (List): dataset from HuggingFace

        Returns:
            List[Dict]: reformatted dataset
        """

        conversations = []
        for _, d in enumerate(data):
            current_conv = d["chosen"]
            split_answer = current_conv.split("Assistant:")

            # take all the list element in split_answer except the last one
            # and joing them with "Assistant:" in a unique string
            previous_convers = split_answer[0]
            for i, s in enumerate(split_answer[1:-1]):
                previous_convers += "Assistant:" + s

            # remove the last characters if they are "\n" from the previous
            # conversation
            previous_convers = previous_convers.rstrip("\n")
            user_input = previous_convers + "\n\n##\n\n"
            completion = "Assistant: " + split_answer[-1]

            conv = {
                "user_input": user_input,
                "completion": completion,
                "score": None,
            }

            conversations.append(conv)
        return conversations

    def save_dataset(
        self, dataset_folder: str, number_of_samples: int, reverse: bool = True
    ) -> None:
        """Save the dataset in the format required by RLHF

        Args:
            dataset_folder (str): path to the folder where the dataset
                will be saved
            number_of_samples (int): number of samples to take from the
                dataset
            reverse (bool, optional): sort the dataset in descending order.
                Defaults to True.
        """

        print("Generate datasets for RLHF")

        # generate actor and reward dataset
        conversations = self.reformat_dataset(self.dataset["train"])
        conversations = self.sort_conversation(conversations, reverse=reverse)

        # save actor training data
        with open(f"{dataset_folder}/actor_training_data.json", "w") as f:
            json.dump(conversations, f, indent=4)

        # sample N number of index from 0 to len(conversations)
        conversations = self.take_n_samples(conversations, number_of_samples)
        conversations = self.sort_conversation(conversations, reverse=reverse)

        # save reward training data
        with open(f"{dataset_folder}/reward_training_data.json", "w") as f:
            json.dump(conversations, f, indent=4)

        # rlhf dataset
        conversations = self.reformat_dataset(self.dataset["test"])

        # sort conversations by length of user_input
        conversations = self.sort_conversation(
            conversations, only_input=True, reverse=reverse
        )

        # save rlhf training data
        with open(f"{dataset_folder}/rlhf_training_data.json", "w") as f:
            json.dump(conversations, f, indent=4)

        print("Generation Completed")



import json
import shutil
import os

import deepspeed
import torch
from accelerate import Accelerator
from beartype import beartype
from beartype.typing import Tuple
from einops import rearrange
from torch.utils.data import DataLoader, Dataset
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
)

from chatllama.rlhf.config import ConfigActor
from chatllama.rlhf.model_list import (
    hf_models_causal_lm,
    llama_models,
    hf_models,
)

from chatllama.rlhf.model_loader import ModelLoader
from chatllama.rlhf.utils import TrainingStats


class ActorModel(torch.nn.Module):
    """Actor model that generates the augmented prompt from the initial
    user_input. The aim is to train this model to generate better prompts.

    Attributes:
        model: The model from LLaMA to be used
        tokenizer: The LLaMA tokenizer
        config (ConfigActor): Configuration for the actor model

    Methods:
        load: Load the model from a path
        save: Save the model to a path
        forward: Compute the action logits for a given sequence.
        generate: Generate a sequence from a given prompt
    """

    def __init__(self, config: ConfigActor) -> None:
        super().__init__()

        # save config
        self.config = config

        # initialize the self.model
        if config.model in llama_models:
            # llama module might not be present when HF models are used
            from chatllama.llama_model import (
                load_model,
                setup_model_parallel,
            )  # noqa

            local_rank, world_size = setup_model_parallel()

            # use load_model_test for testing
            self.model, self.tokenizer = load_model(
                ckpt_dir=config.model_folder,
                tokenizer_path=config.tokenizer_path,
                local_rank=local_rank,
                world_size=world_size,
                froze_embeddings=config.froze_embeddings,
                use_fairscale=config.use_fairscale,
                max_batch_size=config.batch_size,
            )
        elif config.model in hf_models_causal_lm:
            self.tokenizer = self.load_tokenizer(config)
            self.model = AutoModelForCausalLM.from_pretrained(
                config.model,
            )
            self.model.to(config.device)
        else:
            raise ValueError(f"Model {config.model} not supported")

        # load the model from model_folder
        self.load()

    @beartype
    def load(self) -> None:
        """Load the model from the path"""
        # check if there is a model to load
        path = ModelLoader.check_model_path(
            config=self.config,
            is_checkpoint=False,
            current_epoch=None,
        )

        # if there is a model to load
        if path is not None:

            # load the model
            print("Loading ...")
            model_dict = torch.load(path)
            self.model.load_state_dict(model_dict["state_dict"])

    @beartype
    def save(self) -> None:
        """Save the model to the path"""
        # get the path to save the model
        model_folder, model_name, path = ModelLoader.get_model_path(
            config=self.config,
            is_checkpoint=False,
            current_epoch=None,
        )

        # save the model
        print(f"Saving model to {path} ...")
        torch.save(
            {"state_dict": self.model.state_dict()},
            path,
        )

    @staticmethod
    def load_tokenizer(config: ConfigActor):
        """Load the tokenizer from the model name"""
        if config.model in hf_models:
            # load the tokenizer from HF
            tokenizer = AutoTokenizer.from_pretrained(
                config.model,
                padding_side="left",
                padding=True,
                truncation=True,
                model_max_length=config.max_sequence_length,
            )

            # add eos token if not present
            if tokenizer.eos_token is None:
                tokenizer.eos_token = "</s>"
                tokenizer.eos_token_id = 2  # OPT eos-token-id

            # add pad token if not present
            if tokenizer.pad_token is None:
                tokenizer.pad_token = tokenizer.eos_token
                tokenizer.pad_token_id = tokenizer.eos_token_id
        elif config.model in llama_models:

            # llama module might not be present when HF models are used
            from chatllama.llama_model import (
                load_tokenizer,
            )  # noqa

            tokenizer = load_tokenizer()
        return tokenizer

    def parameters(self):
        """Return the parameters of the model"""
        return self.model.parameters()

    @beartype
    def forward(
        self, sequences: torch.Tensor, sequences_mask: torch.Tensor
    ) -> torch.Tensor:
        """Generate logits to have probability distribution over the vocabulary
            of the actions

        Args:
            sequences (torch.Tensor): Sequences of states and actions used to
                    compute token logits for the whole list of sequences
            attention_mask (torch.Tensor): Mask for the sequences attention

        Returns:
            logits (torch.Tensor): Logits for the actions taken
        """
        model_output = self.model.forward(
            sequences, attention_mask=sequences_mask
        )
        # need to return logits for the actions
        if self.config.model in hf_models_causal_lm:
            model_output = model_output.logits
        if self.config.debug:
            print("ActorModel.forward")
            print("model_output_logits shape", model_output.shape)
            print("model_output logits", model_output)
        return model_output

    @beartype
    @torch.no_grad()
    def generate(
        self, states: torch.Tensor, state_mask: torch.Tensor
    ) -> Tuple:
        """Generate actions and sequences=[states, actions] from state
            (i.e. input of the prompt generator model)

        Args:
            state (torch.Tensor): the input of the user
            state_mask (torch.Tensor): Mask for the state input (for padding)

        Returns:
            actions (torch.Tensor): Actions generated from the state
            sequences (torch.Tensor): Sequences generated from the
                state as [states, actions]
        """
        # temperature for the actor
        temperature = self.config.temperature

        # max sequence length for the actor (i.e. prompt + completion)
        max_sequence_length = self.config.max_sequence_length

        # max and min number of tokens to generate
        max_tokens = self.config.max_tokens
        min_tokens = self.config.min_tokens

        # max generation possible given the state and the max sequence length
        max_generation_possible = max_sequence_length - states.shape[1]
        if max_generation_possible < min_tokens:
            raise ValueError(
                f"The prompt is too long w.r.t the "
                f"model sequence length \n"
                f"max_sequence_length={max_sequence_length}\n"
                f"state_length={states.shape[1]}\n"
                f"min_tokens={min_tokens}\n"
                f"max_tokens={max_tokens}\n"
                f"max_generation_possible={max_generation_possible}\n"
            )

        # take the minimum the max_tokens and the max_generation_possible
        max_completion = min(max_tokens, max_generation_possible)

        sequences = self.model.generate(
            input_ids=states,
            attention_mask=state_mask,
            temperature=temperature,
            max_new_tokens=max_completion,
            no_repeat_ngram_size=3,
        )
        actions = sequences[:, states.shape[1] :]  # noqa E203
        if self.config.debug:
            print(
                f"input length {states.shape[1]} \n"
                f"max sequence length {max_sequence_length} \n"
                f"max completion {max_completion} \n"
                f"generated sequence {sequences.shape[1]} \n"
            )
            print("ActorModel.generate")
            print("state", states)
            print("state shape", states.shape)
            print("sequence shape", sequences.shape)
            print("sequence", sequences)
            print("actions shape", actions.shape)
            print("actions", actions)
        return actions, sequences


class ActorDataset(Dataset):
    """Dataset for the pretraining of the actor model
    read a json file with the following format:
    [
        {
            "user_input": "..."
            "completion": "..."
        },
        ...
    ]
    Where:
        user_input: the input of the user
        completion: the output of the user
    """

    def __init__(
        self,
        path: str,
    ) -> None:
        self.path = path
        with open(path, "r") as f:
            data = json.load(f)
        self.data = [d["user_input"] + d["completion"] for d in data]

    def __getitem__(self, idx):
        return self.data[idx]

    def __len__(
        self,
    ):
        return len(self.data)


class ActorTrainer:
    """Used to pre-train the actor model to generate better prompts.

    Args:
        config (ConfigActor): Configuration for the actor model

    Attributes:
        config (ConfigActor): Configuration for the actor model
        model (ActorModel): Actor model
        loss_function (torch.nn.CrossEntropyLoss): Loss function
        optimizer (torch.optim.Adam): Optimizer
        validation_flag (bool): Flag to indicate if the validation dataset
            is provided
        train_dataset (ActorDataset): Training dataset
        train_dataloader (DataLoader): Training dataloader
        validation_dataset (ActorDataset): Validation dataset
        validation_dataloader (DataLoader): Validation dataloader
        scheduler (torch.optim.lr_scheduler): Learning rate scheduler
        training_stats (TrainingStats): Training statistics
        model_engine (ModelEngine): Model engine for deepspeed training
        accelerator (Accelerator): Accelerator for accelerate training

    Methods:
        train: Train the actor model
        load_checkpoint: Load a checkpoint
        save_checkpoint: Save a checkpoint
    """

    def __init__(self, config: ConfigActor) -> None:

        # store config
        self.config = config

        # load the model
        self.actor = ActorModel(config)

        # define loss function
        self.loss_function = torch.nn.CrossEntropyLoss()

        # define optimizer
        self.optimizer = torch.optim.AdamW(
            self.actor.parameters(), lr=config.lr, weight_decay=1e-5
        )

        # check if validation dataset is provided
        self.validation_flag = False
        if config.validation_dataset_path is not None:
            self.validation_flag = True

        # create dataset and dataloaders
        self.train_dataset = ActorDataset(config.train_dataset_path)
        self.train_dataloader = DataLoader(
            self.train_dataset, batch_size=config.batch_size
        )
        if self.validation_flag:
            self.eval_dataset = ActorDataset(config.validation_dataset_path)
            self.validation_dataloader = DataLoader(
                self.eval_dataset, batch_size=config.batch_size
            )

        # define scheduler for the learning rate
        # learning rate is decreased until 10% of the initial value
        self.scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(
            self.optimizer,
            T_0=len(self.train_dataset) // config.batch_size,
            T_mult=1,
            eta_min=config.lr * 0.1,
        )

        # define training statistics
        stat_path = ModelLoader.get_training_stats_path(config)
        self.training_stats = TrainingStats(stat_path)

        # consistency check between accelerate and deepspeed
        if config.accelerate_enable and config.deepspeed_enable:
            raise ValueError(
                "Both DeepSpeed and Accelerate are enabled for the Actor."
                "Please choose one of them."
            )

        # initialize deepspeed
        self.model_engine = None
        if config.deepspeed_enable is True:
            if config.deepspeed_config_path is None:
                raise ValueError(
                    "DeepSpeed config path is None, but deepspeed is enabled"
                )
            if os.path.exists(config.deepspeed_config_path) is False:
                raise ValueError(
                    f"DeepSpeed config path {config.deepspeed_config_path}"
                    f"does not exist"
                )
            (
                self.model_engine,
                self.optimizer,
                self.train_dataloader,
                _,
            ) = deepspeed.initialize(
                args=None,
                model=self.actor,
                model_parameters=self.actor.parameters(),
                training_data=self.train_dataset,
                config=self.config.deepspeed_config_path,
            )
            print("Training with DeepSpeed")

        # initialize accelerate
        self.accelerator = None
        if config.accelerate_enable is True:
            self.accelerator = Accelerator()
            (
                self.actor,
                self.optimizer,
                self.train_dataloader,
                self.scheduler,
            ) = self.accelerator.prepare(
                self.actor,
                self.optimizer,
                self.train_dataloader,
                self.scheduler,
            )
            print("Training with Accelerate")

    @beartype
    def save_checkpoint(
        self,
        current_epoch: int,
        current_step: int,
        max_epochs: int,
        max_steps: int,
    ) -> None:
        """Save the current checkpoint

        Args:
            current_epoch (int): Current epoch
            current_step (int): Current step
            max_epochs (int): Maximum number of epochs
            max_steps (int): Maximum number of steps
        """

        print(
            f"Saving checkpoint for epoch {current_epoch + 1}, "
            f"step {current_step + 1} ..."
        )
        # look for path to save the checkpoint
        model_folder, model_name, path = ModelLoader.get_model_path(
            config=self.config,
            is_checkpoint=True,
            current_epoch=current_epoch,
            current_step=current_step,
            max_epochs=max_epochs,
            max_steps=max_steps,
        )

        # remove the checkpoint if it already exists
        if os.path.exists(path):
            if self.config.deepspeed_enable:
                shutil.rmtree(path)
            else:
                os.remove(path)

        if self.config.deepspeed_enable:
            client_state = {
                "epoch": current_epoch,
                "step": current_step,
            }
            self.model_engine.save_checkpoint(path, client_state=client_state)
        else:
            # save the checkpoint
            torch.save(
                {
                    "state_dict": self.actor.model.state_dict(),
                    "optim_state_dict": self.optimizer.state_dict(),
                    "training_stats": self.training_stats,
                    "epoch": current_epoch,
                    "step": current_step,
                },
                path,
            )

        # remove old checkpoints
        n_checkpoints_to_keep = self.config.n_checkpoints_to_keep
        ModelLoader.delete_old_checkpoints(
            model_folder, model_name, n_checkpoints_to_keep
        )

    @beartype
    def load_checkpoint(
        self,
    ) -> Tuple[int, int]:
        """Load a checkpoint from the model folder

        Returns:
            Tuple[int, int]: Current epoch and current step to resume
                training
        """

        print("Looking for checkpoints...")
        # look for a checkpoint
        path = ModelLoader.check_model_path(
            config=self.config,
            is_checkpoint=True,
            current_epoch=None,
        )

        # if there is a checkpoint
        if path is not None:
            print("Loading ...")

            if self.config.deepspeed_enable:
                # try to load the checkpoint
                try:
                    _, client_state = self.model_engine.load_checkpoint(path)
                except Exception:
                    print(
                        "Checkpoint corrupted!"
                        "Try to remove the last checkpoint."
                        "Now Starting from epoch 0, step 0"
                    )
                    return 0, 0
                # load epoch and step to resume loops
                epoch = client_state["epoch"]
                step = client_state["step"]
            else:
                # try to load the checkpoint
                try:
                    checkpoint = torch.load(path)
                except Exception:
                    print(
                        "Checkpoint corrupted!"
                        "Try to remove the last checkpoint."
                        "Now Starting from epoch 0, step 0"
                    )
                    return 0, 0

                # assing the checkpoint to the model
                epoch = checkpoint["epoch"]
                self.actor.model.load_state_dict(checkpoint["state_dict"])
                self.optimizer.load_state_dict(checkpoint["optim_state_dict"])
                self.trainign_stats = checkpoint["training_stats"]
                step = checkpoint["step"]
                return epoch, step + 1  # return the next episode to train
        return 0, 0

    def add_eos_token(
        self, tokens: torch.Tensor, mask: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        # given tokens and mask, add eos token to the end of each sequence
        # and update the mask
        batch_size, seq_len = tokens.shape
        eos_token = self.actor.tokenizer.eos_token_id

        # see if i can append 1 token
        n_tokens_to_append = min(self.config.max_sequence_length - seq_len, 1)
        n_tokens_to_append = max(n_tokens_to_append, 0)

        # concatenate eos to tokens and mask
        if n_tokens_to_append > 0:
            tokens = torch.cat(
                [
                    tokens,
                    torch.ones(batch_size, n_tokens_to_append)
                    .long()
                    .to(tokens.device)
                    * eos_token,
                ],
                dim=1,
            )
            mask = torch.cat(
                [
                    mask,
                    torch.ones(batch_size, n_tokens_to_append)
                    .long()
                    .to(mask.device),
                ],
                dim=1,
            )
        return tokens, mask

    def train(
        self,
    ) -> None:
        """Train the model"""
        print("Start Actor Model Pretraining")

        # get config parameters
        if self.config.deepspeed_enable:
            batch_size = self.train_dataloader.batch_size
        else:
            batch_size = self.config.batch_size
        epochs = self.config.epochs
        device = self.config.device
        checkpoint_steps = self.config.checkpoint_steps

        # compute the number of iterations
        n_iter = int(len(self.train_dataset) / batch_size)

        # load model_checkpoint
        start_epoch, start_step = self.load_checkpoint()

        if start_epoch == 0 and start_step == 0:
            self.training_stats.clear()

        # counter for the checkpoint
        cnt_checkpoint = 1

        # traing loop
        for epoch in range(start_epoch, epochs):
            self.actor.train()
            for i, input_text in enumerate(self.train_dataloader):

                # skip the first steps if we are resuming training
                if i < start_step:
                    continue

                # tokenize input
                with torch.no_grad():
                    input_tokenized = self.actor.tokenizer(
                        input_text,
                        return_tensors="pt",
                        truncation=True,
                        padding=True,
                    )

                    # split tokens and mask
                    input_tokenized_id = input_tokenized["input_ids"]
                    input_tokenized_mask = input_tokenized["attention_mask"]

                    # add eos token
                    (
                        input_tokenized_id,
                        input_tokenized_mask,
                    ) = self.add_eos_token(
                        input_tokenized_id,
                        input_tokenized_mask,
                    )

                    # split into input and output
                    training_output = input_tokenized_id[:, 1:]
                    training_input = input_tokenized_id[:, :-1]
                    attention_mask = input_tokenized_mask[:, :-1]

                    # move to device
                    training_output = training_output.to(device)
                    training_input = training_input.to(device)
                    attention_mask = attention_mask.to(device)

                # forward pass
                if self.config.deepspeed_enable:
                    est_output = self.model_engine(
                        training_input, attention_mask
                    )
                else:
                    est_output = self.actor(training_input, attention_mask)

                # compute loss
                est_output = rearrange(est_output, "b s v -> (b s) v")
                training_output = rearrange(training_output, "b s -> (b s)")
                loss = self.loss_function(est_output, training_output)
                self.training_stats.training_loss.append(loss.item())

                # backward pass
                if self.config.deepspeed_enable:
                    self.model_engine.backward(loss)
                    self.model_engine.step()
                elif self.config.accelerate_enable:
                    self.optimizer.zero_grad()
                    self.accelerator.backward(loss)
                    self.optimizer.step()
                    self.scheduler.step()
                else:
                    self.optimizer.zero_grad()
                    loss.backward()
                    self.optimizer.step()
                    self.scheduler.step()

                # print progress
                if i % self.config.iteration_per_print == 0:
                    print(
                        f"Epoch: {epoch+1}/{epochs}, "
                        f"Iteration: {i+1}/{n_iter}, "
                        f"Training Loss: {loss}"
                    )

                # save checkpoint periodically
                if cnt_checkpoint % checkpoint_steps == 0:
                    self.save_checkpoint(epoch, i, epochs, n_iter)
                    self.training_stats.save()
                    cnt_checkpoint = 1
                else:
                    cnt_checkpoint += 1

            # Validation
            if self.validation_flag:
                self.actor.eval()
                with torch.no_grad():
                    for i, input_text in enumerate(self.validation_dataloader):

                        # tokenize input
                        input_tokenized = self.actor.tokenizer(
                            input_text, return_tensors="pt", padding=True
                        )
                        validation_output = input_tokenized["input_ids"][:, 1:]
                        validation_input = input_tokenized["input_ids"][:, :-1]
                        attention_mask = input_tokenized["attention_mask"][
                            :, :-1
                        ]

                        # forward pass
                        est_output = self.actor.forward(
                            validation_input, attention_mask
                        )
                        validation_output = rearrange(
                            validation_output, "b s -> (b s)"
                        )

                        # compute loss
                        est_output = rearrange(est_output, "b s v -> (b s) v")
                        loss = self.loss_function(
                            est_output, validation_output
                        )
                        self.training_stats.validation_loss.append(loss.item())

                        # print progress
                        if i % self.config.iteration_per_print == 0:
                            print(
                                f"Epoch: {epoch+1}/{epochs}, "
                                f"Iteration: {i+1}/{n_iter}, "
                                f"Validation Loss: {loss}"
                            )
            # reset start_step after training is resumed
            start_step = 0

        # save the model
        self.actor.save()
        print("Training Finished ")



# llama models
llama_models = ["llama-7B", "llama-13B", "llama-33B", "llama-65B"]

# HF Models
# encoder-decoder models TODO: still not supported
hf_models_seq_2_seq = [
    "google/flan-t5-xxl",
    "google/flan-t5-xl",
    "google/flan-t5-large",
    "google/flan-t5-base",
    "google/flan-t5-small",
]

# decoder only TODO: codegen is still broken
hf_models_causal_lm = [
    "facebook/opt-125m",
    "facebook/opt-1.3b",
    "facebook/opt-2.7b",
    "facebook/opt-6.7b",
    "facebook/opt-11b",
    "facebook/galactica-125m",
    "facebook/galactica-1.3b",
    "facebook/galactica-6.7b",
    "bigscience/bloom-560m",
    "bigscience/bloomz-560m",
    "bigscience/bloom-1b1",
    "bigscience/bloomz-1b1",
    "bigscience/bloom-1b7",
    "bigscience/bloomz-1b7",
    "bigscience/bloom-3b",
    "bigscience/bloomz-3b",
    "bigscience/bloom-7b1",
    "bigscience/bloomz-7b1",
    "EleutherAI/gpt-neo-1.3B",
    "EleutherAI/gpt-neo-1.3B",
    "EleutherAI/gpt-neox-20b",
    "EleutherAI/gpt-j-6B" "gpt2",
    "gpt2-large",
    "gpt2-xl",
    "Salesforce/codegen-350M-mono",
    "Salesforce/codegen-2B-mono",
    "Salesforce/codegen-6B-mono",
    "Salesforce/codegen-16B-mono",
]

# create a list of all the models from hf
hf_models = hf_models_seq_2_seq + hf_models_causal_lm



import json
import shutil
import os

import deepspeed
import torch
from accelerate import Accelerator
from beartype import beartype
from beartype.typing import Iterable, Tuple
from einops.layers.torch import Rearrange
from torch.utils.data import Dataset, DataLoader
from transformers import (
    AutoModel,
    AutoTokenizer,
)

from chatllama.rlhf.config import ConfigReward
from chatllama.rlhf.model_list import hf_models
from chatllama.rlhf.model_loader import ModelLoader
from chatllama.rlhf.utils import TrainingStats


class RewardModel(torch.nn.Module):
    """Model to be trained to predict the reward for RL.
    or to be used as Critic in RL. It is a Language Model with a head
    that predicts the reward (a scalar) for a given sequence of tokens.

    Attributes:
        model (torch.nn.Module): Model to be used for the reward model
        tokenizer (torch.nn.Module): Tokenizer to be used for the reward model
        head (torch.nn.Module): Head to be used for the reward model
        config (ConfigReward): Config parameters for the reward model

    Methods:
        load_tokenizer: Load the tokenizer for the reward model
        forward: Forward pass of the model (used by the critic)
        save: Save the model
        load: Load the model
        get_reward: Get the reward for a given input (used by the reward model)
        parameters: Return the parameters of the reward model

    """

    def __init__(self, config: ConfigReward) -> None:
        super().__init__()

        # store config
        self.config = config

        # initialize the self.model
        head_hidden_size = config.model_head_hidden_size
        if config.model in hf_models:
            self.tokenizer = self.load_tokenizer(config)
            self.model = AutoModel.from_pretrained(config.model)
            head_dim = self.model.config.hidden_size
            if config.model.startswith("gpt2"):
                head_dim = self.model.config.n_embd
            self.head = torch.nn.Sequential(
                torch.nn.Linear(head_dim, head_hidden_size),
                torch.nn.ReLU(),
                torch.nn.Linear(head_hidden_size, 1),
                Rearrange("... 1 -> ..."),
            )
        else:
            raise ValueError(f"Model {config.model} not supported")

        # load the model
        self.load()

        # freeze model parameters (only train the head)
        # for param in self.model.parameters():
        #     param.requires_grad = False

        # move model to device
        self.model.to(config.device)
        self.head.to(config.device)

    @staticmethod
    def load_tokenizer(config: ConfigReward):
        # load tokenizer from HF
        tokenizer = AutoTokenizer.from_pretrained(
            config.model,
            padding_side="left",
            padding=True,
            truncation=True,
            model_max_length=config.max_sequence_length,
        )

        # add eos token if not present
        if tokenizer.eos_token is None:
            tokenizer.eos_token = "</s>"
            tokenizer.eos_token_id = 2  # OPT  eos token id

        # add pad token if not present
        if tokenizer.pad_token is None:
            tokenizer.pad_token = tokenizer.eos_token
            tokenizer.pad_token_id = tokenizer.eos_token_id
        return tokenizer

    @beartype
    def load(self) -> None:
        """Load the model from the path"""
        # look for a pretrained model
        path = ModelLoader.check_model_path(
            config=self.config,
            is_checkpoint=False,
            current_epoch=None,
        )

        # check if the model exists
        if path is not None:

            # load the model from the path
            print("Loading ...")
            model_dict = torch.load(path)
            self.model.load_state_dict(model_dict["model"])
            self.head.load_state_dict(model_dict["head"])

    @beartype
    def save(self) -> None:
        """Save the model to the path"""
        # get the path to save the model
        model_folder, model_name, path = ModelLoader.get_model_path(
            config=self.config,
            is_checkpoint=False,
            current_epoch=None,
        )

        # save the model
        print(f"Saving model to {path} ...")
        torch.save(
            {"model": self.model.state_dict(), "head": self.head.state_dict()},
            path,
        )

    @beartype
    def parameters(
        self,
    ) -> Iterable[torch.nn.Parameter]:
        """Return the parameters of the reward model"""
        for p in self.model.parameters():
            yield p
        for p in self.head.parameters():
            yield p

    @beartype
    def forward(
        self, output_sequence: torch.Tensor, output_sequence_mask: torch.Tensor
    ) -> torch.Tensor:
        """Generate the sequence of rewards for the given output sequence
        what is the quality of the output sequence tokens?

        Args:
            output_sequence (torch.Tensor): The sequence of tokens to be
                evaluated
            output_sequence_mask (torch.Tensor): Mask for the attention

        Returns:
            torch.Tensor: Rewards for the given output sequence
        """
        output = self.model(
            output_sequence, attention_mask=output_sequence_mask
        )

        # What if the output_sequence is longer than the max context of
        # the model?
        rewards = self.head(output.last_hidden_state)
        if self.config.debug:
            print("RewardModel.forward")
            print("output_sequence.shape", output_sequence.shape)
            print("output_sequence", output_sequence)
            print("reward.shape", rewards.shape)
            print("reward", rewards)
        return rewards

    @beartype
    def get_reward(
        self, output_sequence: torch.Tensor, output_sequence_mask: torch.Tensor
    ) -> torch.Tensor:
        """Get the reward for the given output sequence

        Args:
            output_sequence (torch.Tensor): The concatenation of initial input
                and actor output as tokens
            output_sequence_mask (torch.Tensor): Mask for the attention
        """
        if output_sequence.shape[1] > self.config.max_sequence_length:
            raise ValueError(
                f"Output sequence is too long: {output_sequence.shape[1]}"
                f" > {self.config.max_sequence_length}"
            )
        rewards = self.forward(output_sequence, output_sequence_mask)
        return rewards[:, -1]


# just to keep namings consistent
CriticModel = RewardModel


class RewardDataset(Dataset):
    """Dataset class for the reward model
    read a json file with the following format:
    [
        {
            "user_input": "...",
            "completion": "...",
            "score": ...
        },
        ...
    ]
    Where:
        user_input: the initial input of the user
        completion: the completion generated by the model
        score: the score given by the user to the completion (or by the LLM)
    """

    def __init__(self, path: str) -> None:
        print(f"Loading dataset from {path}")
        with open(path, "r") as f:
            self.data = list(json.load(f))
        print(f"Loaded {len(self.data)} samples")

    def __getitem__(self, idx: int):
        user_input = self.data[idx]["user_input"]
        completion = self.data[idx]["completion"]
        score = float(self.data[idx]["score"])
        item = (user_input + completion, score)
        return item

    def __len__(
        self,
    ):
        return len(self.data)


class RewardTrainer:
    """Class to train the reward model

    Args:
        config (ConfigModel): Config parameters for the model

    Attributes:
        model (RewardModel): Reward model
        config (ConfigModel): Config parameters for the model
        optimizer (torch.optim): Optimizer for the model
        loss_function (torch.nn): Loss function for the model
        validation_flag (bool): Flag to indicate if the validation dataset
            is available
        train_dataset (RewardDataset): Dataset for training
        validation_dataset (RewardDataset): Dataset for validation
        train_dataloader (DataLoader): Dataloader for training
        validation_dataloader (DataLoader): Dataloader for validation
        scheduler (torch.optim.lr_scheduler): Scheduler for the optimizer
        training_stats (List[Dict]): List of dictionaries with the training
            statistics
        model_engine (ModelEngine): Model engine to train the model
            using deepspeed
        accelerator (Accelerator): Accelerator to train the model using
            accelerate by HF.


    Methods:
        train: Train the reward model
        save_checkpoints: Save the checkpoints of the model
        load_checkpoints: Load the checkpoints of the model
    """

    def __init__(self, config: ConfigReward) -> None:

        # save the config
        self.config = config

        # load the model
        self.reward = RewardModel(config)

        # optimizer
        self.optimizer = torch.optim.AdamW(
            self.reward.parameters(), lr=config.lr
        )

        # loss function
        self.loss_function = torch.nn.MSELoss()

        # check validation dataset
        self.validation_flag = False
        if config.validation_dataset_path is not None:
            self.validation_flag = True

        # create dataset and dataloaders
        self.train_dataset = RewardDataset(config.train_dataset_path)
        self.train_dataloader = DataLoader(
            self.train_dataset, batch_size=config.batch_size
        )
        if self.validation_flag:
            self.eval_dataset = RewardDataset(config.validation_dataset_path)
            self.validation_dataloader = DataLoader(
                self.eval_dataset, batch_size=config.batch_size
            )

        # intilize scheduler - learning rate will drop to 10% of the initial
        # value
        self.scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(
            self.optimizer,
            T_0=len(self.train_dataset) // config.batch_size,
            T_mult=1,
            eta_min=config.lr * 0.1,
            last_epoch=-1,
        )

        # initialize training stats
        stats_path = ModelLoader.get_training_stats_path(config)
        self.training_stats = TrainingStats(stats_path)

        # consistency check between accelerate and deepspeed
        if config.accelerate_enable and config.deepspeed_enable:
            raise ValueError(
                "Both DeepSpeed and Accelerate are enabled for the Reward."
                "Please choose one of them."
            )

        # initialize deepspeed
        self.model_engine = None
        if config.deepspeed_enable is True:

            if config.deepspeed_config_path is None:
                raise ValueError(
                    "DeepSpeed config path is None, but deepspeed is enabled"
                )
            if os.path.exists(config.deepspeed_config_path) is False:
                raise ValueError(
                    f"DeepSpeed config path {config.deepspeed_config_path}"
                    f"does not exist"
                )
            (
                self.model_engine,
                self.optimizer,
                self.train_dataloader,
                self.scheduler,
            ) = deepspeed.initialize(
                args=None,
                model=self.reward,
                model_parameters=self.reward.parameters(),
                training_data=self.train_dataset,
                config=self.config.deepspeed_config_path,
            )
            print("Training with DeepSpeed")

        # initialize accelerate
        self.accelerator = None
        if config.accelerate_enable is True:
            self.accelerator = Accelerator()
            (
                self.reward,
                self.optimizer,
                self.train_dataloader,
                self.scheduler,
            ) = self.accelerator.prepare(
                self.reward,
                self.optimizer,
                self.train_dataloader,
                self.scheduler,
            )
            print("Training with Accelerate")

    @beartype
    def save_checkpoint(
        self,
        current_epoch: int,
        current_step: int,
        max_epochs: int,
        max_steps: int,
    ) -> None:
        """Save the checkpoints of the model

        Args:
            current_epoch (int): Current epoch
            current_step (int): Current step
            max_epochs (int): Maximum number of epochs
            max_steps (int): Maximum number of steps
        """

        print(
            f"Saving checkpoint for epoch {current_epoch + 1}, "
            f" step {current_step} ..."
        )

        # get the path to save the checkpoint
        model_folder, model_name, path = ModelLoader.get_model_path(
            config=self.config,
            is_checkpoint=True,
            current_epoch=current_epoch,
            current_step=current_step,
            max_epochs=max_epochs,
            max_steps=max_steps,
        )

        # remove the checkpoint if it already exists
        if os.path.exists(path):
            if self.config.deepspeed_enable:
                shutil.rmtree(path)
            else:
                os.remove(path)

        # save the checkpoint
        if self.config.deepspeed_enable:
            client_state = {
                "epoch": current_epoch,
                "step": current_step,
            }
            self.model_engine.save_checkpoint(path, client_state=client_state)
        else:
            torch.save(
                {
                    "state_dict": self.reward.model.state_dict(),
                    "optim_state_dict": self.optimizer.state_dict(),
                    "scheduler_state_dict": self.scheduler.state_dict(),
                    "training_stats": self.training_stats,
                    "epoch": current_epoch,
                    "step": current_step,
                },
                path,
            )

    @beartype
    def load_checkpoint(
        self,
    ) -> Tuple[int, int]:
        """Load the checkpoints of the model

        Returns:
            Tuple[int, int]: The current epoch and step
                from which you should resume the training
        """

        print("Looking for checkpoints...")
        # look for the checkpoints
        path = ModelLoader.check_model_path(
            config=self.config,
            is_checkpoint=True,
            current_epoch=None,
        )

        # check if a checkpoint exists
        if path is not None:
            print("Loading ...")

            if self.config.deepspeed_enable:
                # try to load the checkpoint
                try:
                    _, client_state = self.model_engine.load_checkpoint(path)
                except Exception:
                    print(
                        "Checkpoint corrupted!"
                        "Try to remove the last checkpoint."
                        "Now Starting from epoch 0, step 0"
                    )
                    return 0, 0
                # load epoch and step to resume loops
                epoch = client_state["epoch"]
                step = client_state["step"]
            else:
                # try to load the checkpoint
                try:
                    checkpoint = torch.load(path)
                except Exception:
                    print(
                        "Checkpoint corrupted!"
                        "Try to remove the last checkpoint."
                        "Now Starting from epoch 0, step 0"
                    )
                    return 0, 0

                # load the model parameters and optimizer parameters
                # from the checkpoint
                epoch = checkpoint["epoch"]
                self.reward.model.load_state_dict(checkpoint["state_dict"])
                self.optimizer.load_state_dict(checkpoint["optim_state_dict"])
                self.scheduler.load_state_dict(
                    checkpoint["scheduler_state_dict"]
                )
                self.training_stats = checkpoint["training_stats"]
                step = checkpoint["step"]
            return epoch, step + 1  # return the next episode to train
        return 0, 0

    def train(
        self,
    ) -> None:
        """Train the reward model"""
        print("Start Training the Reward Model")

        # get config parameters
        if self.config.deepspeed_enable:
            batch_size = self.train_dataloader.batch_size
        else:
            batch_size = self.config.batch_size
        epochs = self.config.epochs
        device = self.config.device
        iteration_per_print = self.config.iteration_per_print
        checkpoint_steps = self.config.checkpoint_steps

        # compute the number of iterations
        n_iter = int(len(self.train_dataset) / batch_size)

        # load checkpoint
        start_epoch, start_step = self.load_checkpoint()

        # counter for the checkpoint
        cnt_checkpoints = 1

        # traing loop
        for epoch in range(start_epoch, epochs):
            self.reward.train()
            for i, inputs in enumerate(self.train_dataloader):

                # skip the steps if resuming from a checkpoint
                if i < start_step:
                    continue

                # get the inputs
                input_text = inputs[0]
                score = inputs[1]

                # tokenize the input
                with torch.no_grad():
                    input_tokens = self.reward.tokenizer(
                        input_text,
                        return_tensors="pt",
                        truncation=True,
                        padding=True,
                    )
                    output = torch.as_tensor(
                        score, dtype=torch.float32, device=device
                    )

                # forward pass
                if self.config.deepspeed_enable:
                    est_output = self.model_engine(
                        input_tokens["input_ids"].to(device),
                        input_tokens["attention_mask"].to(device),
                    )[:, -1]
                else:
                    est_output = self.reward.get_reward(
                        input_tokens["input_ids"].to(device),
                        input_tokens["attention_mask"].to(device),
                    )

                # compute the loss
                loss = self.loss_function(est_output, output)
                self.training_stats.training_loss.append(loss.item())

                # backward pass
                if self.config.deepspeed_enable:
                    self.model_engine.backward(loss)
                    self.model_engine.step()
                elif self.config.accelerate_enable:
                    self.optimizer.zero_grad()
                    self.accelerator.backward(loss)
                    self.optimizer.step()
                    self.scheduler.step()
                else:
                    self.optimizer.zero_grad()
                    loss.backward()
                    self.optimizer.step()
                    self.scheduler.step()

                # print progress
                if i % iteration_per_print == 0:
                    print(
                        f"Epoch: {epoch+1}/{epochs}, "
                        f"Iteration: {i+1}/{n_iter}, "
                        f"Training Loss: {loss.item()}"
                    )
                    printed_est_output = [
                        round(float(x), 1) for x in est_output.cpu().tolist()
                    ]
                    print(
                        "prediction",
                        printed_est_output,
                        "target",
                        score.cpu().tolist(),
                    )

                # checkpoints saving
                if cnt_checkpoints % checkpoint_steps == 0:
                    self.save_checkpoint(epoch, i, epochs, n_iter)
                    cnt_checkpoints = 1
                else:
                    cnt_checkpoints += 1

            # Validation
            if self.validation_flag:
                self.reward.eval()
                with torch.no_grad():
                    for i, (text, score) in enumerate(
                        self.validation_dataloader
                    ):

                        # tokenize inputs
                        input_tokens = self.reward.tokenizer(
                            text, return_tensors="pt", padding=True
                        )
                        input_tokens = input_tokens.to(device)
                        # TODO: check on the length of the input tokens if
                        # they are too many it can create problems
                        output = torch.tensor(score, dtype=torch.float32).to(
                            device
                        )

                        # forward pass
                        est_output = self.reward.get_reward(
                            input_tokens["input_ids"],
                            input_tokens["attention_mask"],
                        )

                        # compute loss
                        loss = self.loss_function(est_output, output)
                        self.training_stats.validation_loss.append(loss.item())

                        # print progress
                        if i % iteration_per_print == 0:
                            print(
                                f"Epoch: {epoch+1}/{epochs}, "
                                f"Iteration: {i+1}/{n_iter}, "
                                f"Validation Loss: {loss.item()}"
                            )
            # reset start_step after training is resumed
            start_step = 0

        # save the model at the end of the training
        self.reward.save()



REWARD_TEMPLATE = dict(
    template=(
        "You have to evaluate the following chat with a score between 0 and 5"
        "You MUST evaluate: text quality, content quality and"
        "coherence.\n"
        "You MUST return only the number that represents your"
        "judgment.\n"
        "The assignement is:\n{user_input}\n"
        "The completion is:\n{completion}\n"
    ),
    input_variables=["user_input", "completion"],
)


AI_CHATBOT_TEMPLATE = dict(
    template=(
        "Assistant is a large language model trained by Meta and Nebuly.ai\n"
        "Assistant is designed to be able to assist with a wide range of "
        "tasks, from answering simple questions to providing in-depth "
        "explanations and discussions on a wide range of topics. As a "
        "language model, Assistant is able to generate human-like text "
        "based on the input it receives, allowing it to engage in "
        "natural-sounding conversations and provide responses that are "
        "coherent and relevant to the topic at hand.\n\n"
        "Assistant is constantly learning and improving, and its capabilities "
        "are constantly evolving. It is able to process and understand large "
        "amounts of text, and can use this knowledge to provide accurate and "
        "informative responses to a wide range of questions. Additionally, "
        "Assistant is able to generate its own text based on the input it "
        "receives, allowing it to engage in discussions and provide "
        "explanations and descriptions on a wide range of topics.\n\n"
        "Overall, Assistant is a powerful tool that can help with a wide "
        "range of tasks and provide valuable insights and information on a "
        "wide range of topics. Whether you need help with a specific "
        "question or just want to have a conversation about a particular "
        "topic, Assistant is here to assist.\n\n{history}\n\n"
        "Human: {human_input}\n"
        "Assistant:"
    ),
    input_variables=["history", "human_input"],
)


PERSON_CHATBOT_TEMPLATE = dict(
    template=(
        "You are a human chatting with a chatbot. The chatbot is a large "
        "language model trained by Meta and Nebuly-ai\n"
        "The chatbot is designed to be able to assist you with a wide range "
        "of tasks, from answering simple questions to providing in-depth "
        "explanations and discussions on a wide range of topics. You are a "
        "human and you are testing the chatbot. Ask the chatbot questions and"
        "see how it responds. You can also ask the chatbot to tell you a "
        "story."
        "\n\n{history}\n\n"
        "Chatbot: {chatbot_input}\n"
        "Human:"
    ),
    input_variables=["history", "chatbot_input"],
)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This software may be used and distributed according to the terms
# of the GNU General Public License version 3.

import json
import math
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Tuple, List, Union, Optional

import deepspeed
import torch
import torch.distributed
import torch.nn.functional as F
import fairscale.nn.model_parallel.initialize as fs_init
from fairscale.nn.model_parallel.initialize import initialize_model_parallel
from fairscale.nn.model_parallel.layers import (
    ParallelEmbedding,
    RowParallelLinear,
    ColumnParallelLinear,
)
from torch import nn
from transformers import AutoTokenizer

from llama import Tokenizer
from llama.generation import sample_top_p


class MyTokenizer:
    """Masked tokenizer of hugging face to be similar to the one of meta,
    just used for testing purposes.
    """

    def __init__(self, model_path: Optional[str] = None):

        if model_path is None:
            self.sp_model = AutoTokenizer.from_pretrained("gpt2")
        else:
            self.sp_model = AutoTokenizer.from_pretrained(model_path)

        self.n_words = self.sp_model.vocab_size
        self.bos_id = self.sp_model.bos_token_id
        self.eos_id = self.sp_model.eos_token_id
        self.pad_id = self.sp_model.eos_token_id

    def encode(
        self,
        s: str,
        bos: bool = True,
        eos: bool = True,
        truncation: bool = True,
    ) -> List[int]:
        output = self.sp_model.encode(s, truncation=truncation)
        t = list(output)
        if bos:
            t = [self.bos_id] + t
        if eos:
            t = t + [self.eos_id]
        return t

    def decode(self, t: List[int]) -> str:
        input = torch.as_tensor(t)
        output = self.sp_model.decode(input)
        return output


class HFLikeTokenizer:
    def __init__(self, tokenizer: Tokenizer):
        self.tokenizer = tokenizer

        # assign attributes from real tokenizer to masked one
        self.pad_id = self.tokenizer.pad_id
        self.eos_id = self.tokenizer.eos_id
        self.bos_id = self.tokenizer.bos_id

        # mask attribute to be similar to hugging face
        self.eos_token_id = self.tokenizer.eos_id
        self.pad_token_id = self.tokenizer.pad_id

        # to match hugging face attribute
        self.pad_token_id = self.pad_id

    def create_sequence_mask(self, tokens: torch.Tensor) -> torch.Tensor:
        mask = torch.where(
            tokens == self.tokenizer.pad_id,
            torch.zeros_like(tokens),
            torch.ones_like(tokens),
        )
        mask = torch.where(
            tokens == self.tokenizer.bos_id, torch.zeros_like(tokens), mask
        )
        mask = torch.where(
            tokens == self.tokenizer.eos_id, torch.zeros_like(tokens), mask
        )
        return mask

    def __call__(self, texts: Union[List[str], str], *args, **kwargs):
        if isinstance(texts, str):
            text = self.tokenizer.encode(texts, bos=True, eos=True)
            tokens = torch.tensor(text).long()
            mask = torch.ones_like(tokens)
        else:
            texts = [
                self.tokenizer.encode(text, bos=True, eos=True)
                for text in texts
            ]
            max_len = max(len(text) for text in texts)
            tokens = torch.full(
                (len(texts), max_len), self.tokenizer.pad_id
            ).long()
            for i, text in enumerate(texts):
                tokens[i, -len(text) :] = torch.tensor(  # noqa E203
                    text
                ).long()

            # TODO: decide how eos and bos should be handled - i need to mask
            # them? or not?
            mask = self.create_sequence_mask(tokens)
            for i in range(tokens.shape[0]):
                current_tokens = tokens[i, mask[i] == 1]
                tokens[
                    i, -len(current_tokens) - 1 : -1  # noqa E203
                ] = current_tokens
            mask = self.create_sequence_mask(tokens)

            # convert `pad_id` from -1 to 0, otherwise embedding will cause out
            # of bounds.
            tokens = torch.where(
                tokens == self.tokenizer.pad_id,
                torch.zeros_like(tokens),
                tokens,
            )
        output = {
            "input_ids": tokens,
            "attention_mask": mask,
        }
        return output

    def decode(self, tokens):
        return self.tokenizer.decode(tokens)


@dataclass
class ModelArgs:
    """This class is a modification of the ModelArgs class implemented in
    the LLaMA repo. The class has been modified for training, since the
    original one just supports inference.
    """

    dim: int = 512
    n_layers: int = 8
    n_heads: int = 8
    # defined later by tokenizer
    vocab_size: int = -1
    # make SwiGLU hidden layer size multiple of large power of 2
    multiple_of: int = 256
    norm_eps: float = 1e-5

    max_batch_size: int = 32
    max_seq_len: int = 1024

    # added attributes
    froze_embeddings: bool = True
    use_fairscale: bool = True


class RMSNorm(torch.nn.Module):
    """This class is a modification of the RMSNorm class implemented in
    the LLaMA repo. The class has been modified for training, since the
    original one just supports inference.
    """

    def __init__(self, dim: int, eps: float = 1e-6):
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(dim))

    def _norm(self, x):
        return x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps)

    def forward(self, x):
        output = self._norm(x.float()).type_as(x)
        return output * self.weight


def precompute_freqs_cis(dim: int, end: int, theta: float = 10000.0):
    freqs = 1.0 / (
        theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim)
    )
    t = torch.arange(end, device=freqs.device)  # type: ignore
    freqs = torch.outer(t, freqs).float()  # type: ignore
    freqs_cis = torch.polar(torch.ones_like(freqs), freqs)  # complex64
    return freqs_cis


def reshape_for_broadcast(freqs_cis: torch.Tensor, x: torch.Tensor):
    ndim = x.ndim
    assert 0 <= 1 < ndim
    assert freqs_cis.shape == (x.shape[1], x.shape[-1])
    shape = [
        d if i == 1 or i == ndim - 1 else 1 for i, d in enumerate(x.shape)
    ]
    return freqs_cis.view(*shape)


def apply_rotary_emb(
    xq: torch.Tensor,
    xk: torch.Tensor,
    freqs_cis: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    xq_ = torch.view_as_complex(xq.float().reshape(*xq.shape[:-1], -1, 2))
    xk_ = torch.view_as_complex(xk.float().reshape(*xk.shape[:-1], -1, 2))
    freqs_cis = reshape_for_broadcast(freqs_cis, xq_)
    xq_out = torch.view_as_real(xq_ * freqs_cis).flatten(3)
    xk_out = torch.view_as_real(xk_ * freqs_cis).flatten(3)
    return xq_out.type_as(xq), xk_out.type_as(xk)


class Attention(nn.Module):
    """This class is a modification of the Attention class implemented in
    the LLaMA repo. The class has been modified for training, since the
    original one just supports inference.
    """

    def __init__(self, args: ModelArgs):
        super().__init__()

        if args.use_fairscale:
            self.n_local_heads = (
                args.n_heads // fs_init.get_model_parallel_world_size()
            )
        else:
            self.n_local_heads = args.n_heads
        self.head_dim = args.dim // args.n_heads

        if args.use_fairscale:
            self.wq = ColumnParallelLinear(
                args.dim,
                args.n_heads * self.head_dim,
                bias=False,
                gather_output=False,
                init_method=lambda x: x,
            )
            self.wk = ColumnParallelLinear(
                args.dim,
                args.n_heads * self.head_dim,
                bias=False,
                gather_output=False,
                init_method=lambda x: x,
            )
            self.wv = ColumnParallelLinear(
                args.dim,
                args.n_heads * self.head_dim,
                bias=False,
                gather_output=False,
                init_method=lambda x: x,
            )
            self.wo = RowParallelLinear(
                args.n_heads * self.head_dim,
                args.dim,
                bias=False,
                input_is_parallel=True,
                init_method=lambda x: x,
            )
        else:
            self.wq = nn.Linear(
                args.dim, args.n_heads * self.head_dim, bias=False
            )
            self.wk = nn.Linear(
                args.dim, args.n_heads * self.head_dim, bias=False
            )
            self.wv = nn.Linear(
                args.dim, args.n_heads * self.head_dim, bias=False
            )
            self.wo = nn.Linear(
                args.n_heads * self.head_dim, args.dim, bias=False
            )

        self.dim_cache = (
            args.max_batch_size,
            args.max_seq_len,
            self.n_local_heads,
            self.head_dim,
        )
        self.cache_k = torch.zeros(self.dim_cache).cuda()

        self.cache_v = torch.zeros(self.dim_cache).cuda()

    def forward(
        self,
        x: torch.Tensor,
        kv_mask: torch.Tensor,
        freqs_cis: torch.Tensor,
        cache_k: Optional[torch.Tensor] = None,
        cache_v: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        start_pos = 0  # Temporary

        bsz, seqlen, _ = x.shape
        xq, xk, xv = self.wq(x), self.wk(x), self.wv(x)

        xq = xq.view(bsz, seqlen, self.n_local_heads, self.head_dim)
        xk = xk.view(bsz, seqlen, self.n_local_heads, self.head_dim)
        xv = xv.view(bsz, seqlen, self.n_local_heads, self.head_dim)

        xq, xk = apply_rotary_emb(xq, xk, freqs_cis=freqs_cis)

        # Modified code to allow training, caching is not good for training
        if (cache_k is None and cache_v is not None) or (
            cache_k is not None and cache_v is None
        ):
            raise ValueError("cache_k is None while cache_v is not None")
        if cache_k is None:
            keys = xk
            values = xv
        else:
            cache_k.to(xk.device)
            cache_v.to(xv.device)
            cache_k[:bsz, start_pos : start_pos + seqlen] = xk  # noqa E203
            cache_v[:bsz, start_pos : start_pos + seqlen] = xv  # noqa E203
            keys = self.cache_k[:bsz, : start_pos + seqlen]  # noqa E203
            values = self.cache_v[:bsz, : start_pos + seqlen]  # noqa E203

        xq = xq.transpose(1, 2)
        keys = keys.transpose(1, 2)
        values = values.transpose(1, 2)
        scores = torch.matmul(xq, keys.transpose(2, 3)) / math.sqrt(
            self.head_dim
        )
        if kv_mask is not None:
            scores = scores + kv_mask
        scores = F.softmax(scores.float(), dim=-1).type_as(xq)
        output = torch.matmul(scores, values)
        output = output.transpose(1, 2).contiguous().view(bsz, seqlen, -1)
        if cache_k is None:
            return self.wo(output), None, None
        else:
            return self.wo(output), self.cache_k, self.cache_v


class FeedForward(nn.Module):
    """This class is a modification of the FeedForward class implemented in
    the LLaMA repo. The class has been modified for training, since the
    original one just supports inference.
    """

    def __init__(
        self, dim: int, hidden_dim: int, multiple_of: int, use_fairscale: bool
    ):
        super().__init__()
        hidden_dim = int(2 * hidden_dim / 3)
        hidden_dim = multiple_of * (
            (hidden_dim + multiple_of - 1) // multiple_of
        )

        if use_fairscale:
            self.w1 = ColumnParallelLinear(
                dim,
                hidden_dim,
                bias=False,
                gather_output=False,
                init_method=lambda x: x,
            )
            self.w2 = RowParallelLinear(
                hidden_dim,
                dim,
                bias=False,
                input_is_parallel=True,
                init_method=lambda x: x,
            )
            self.w3 = ColumnParallelLinear(
                dim,
                hidden_dim,
                bias=False,
                gather_output=False,
                init_method=lambda x: x,
            )
        else:
            self.w1 = nn.Linear(dim, hidden_dim, bias=False)
            self.w2 = nn.Linear(hidden_dim, dim, bias=False)
            self.w3 = nn.Linear(dim, hidden_dim, bias=False)

    def forward(self, x):
        return self.w2(F.silu(self.w1(x)) * self.w3(x))


class TransformerBlock(nn.Module):
    """This class is a modification of the TransformerBlock class
    implemented in the LLaMA repo. The class has been modified for training,
    since the original one just supports inference.
    """

    def __init__(self, layer_id: int, args: ModelArgs):
        super().__init__()
        self.n_heads = args.n_heads
        self.dim = args.dim
        self.head_dim = args.dim // args.n_heads
        self.attention = Attention(args)
        self.feed_forward = FeedForward(
            dim=args.dim,
            hidden_dim=4 * args.dim,
            multiple_of=args.multiple_of,
            use_fairscale=args.use_fairscale,
        )
        self.layer_id = layer_id
        self.attention_norm = RMSNorm(args.dim, eps=args.norm_eps)
        self.ffn_norm = RMSNorm(args.dim, eps=args.norm_eps)
        self.use_fairscale = args.use_fairscale

    def forward(
        self,
        x: torch.Tensor,
        attention_mask: torch.Tensor,
        freqs_cis: torch.Tensor,
        cache_k: Optional[torch.Tensor] = None,
        cache_v: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[torch.Tensor]]:
        # modified from orignal code to enable external cache
        attention_mask = attention_mask[:, None, :, :]
        if self.use_fairscale:
            attention_mask = attention_mask.expand(
                -1,
                self.n_heads // fs_init.get_model_parallel_world_size(),
                -1,
                -1,
            )
        else:
            attention_mask = attention_mask.expand(-1, self.n_heads, -1, -1)
        attn, cache_k, cache_v = self.attention.forward(
            self.attention_norm(x), attention_mask, freqs_cis, cache_k, cache_v
        )
        h = x + attn
        out = h + self.feed_forward.forward(self.ffn_norm(h))
        return out, cache_k, cache_v


class Transformer(nn.Module):
    """This class is a modification of the Transformer class implemented in
    the LLaMA repo. The class has been modified for training, since the
    original one just supports inference. The generate method was inspired by
    the generate function you can find in `llama.generation`.
    """

    def __init__(self, params: ModelArgs):
        super().__init__()

        self.params = params
        self.vocab_size = params.vocab_size
        self.n_layers = params.n_layers
        if params.use_fairscale:
            self.n_local_heads = (
                params.n_heads // fs_init.get_model_parallel_world_size()
            )
        else:
            self.n_local_heads = params.n_heads

        self.head_dim = params.dim // params.n_heads
        dim = (
            params.max_batch_size,
            params.max_seq_len,
            self.n_local_heads,
            self.head_dim,
        )
        self.cache_k = [torch.zeros(dim) for _ in range(self.n_layers)]
        self.cache_v = [torch.zeros(dim) for _ in range(self.n_layers)]

        if params.use_fairscale:
            self.tok_embeddings = ParallelEmbedding(
                params.vocab_size, params.dim, init_method=lambda x: x
            )
        else:
            self.tok_embeddings = nn.Embedding(params.vocab_size, params.dim)

        if params.froze_embeddings:
            for param in self.tok_embeddings.parameters():
                param.requires_grad = False

        self.layers = torch.nn.ModuleList()
        for layer_id in range(params.n_layers):
            self.layers.append(TransformerBlock(layer_id, params))

        self.norm = RMSNorm(params.dim, eps=params.norm_eps)
        if params.use_fairscale:
            self.output = ColumnParallelLinear(
                params.dim,
                params.vocab_size,
                bias=False,
                init_method=lambda x: x,
            )
        else:
            self.output = nn.Linear(params.dim, params.vocab_size, bias=False)

        # TODO: How too modify this for training?
        self.freqs_cis = precompute_freqs_cis(
            self.params.dim // self.params.n_heads, self.params.max_seq_len * 2
        )

    def forward(
        self, tokens: torch.Tensor, attention_mask: torch.Tensor
    ) -> torch.Tensor:
        attention_mask = attention_mask.detach()
        logits = self._forward(tokens, attention_mask)
        return logits

    def _forward(
        self, tokens: torch.Tensor, attention_mask: torch.Tensor
    ) -> torch.Tensor:
        _bsz, seqlen = tokens.shape
        h = self.tok_embeddings(tokens)
        self.freqs_cis = self.freqs_cis.to(h.device)
        # TEMPORARY FIX, need to understand how to manage the positioning
        # embedding and the batch size with the current padding and masking.
        start_pos = 1
        freqs_cis = self.freqs_cis[start_pos : start_pos + seqlen]  # noqa E203
        # mask has size (bsz, seqlen). It should be transformed in
        # (bsz, seqlen, seqlen)
        # if the mask is a boolean tensor, convert it to int
        if attention_mask.dtype == torch.bool:
            attention_mask = attention_mask.long()
        kv_mask = attention_mask[:, None, :].expand(_bsz, seqlen, seqlen)
        kv_mask = torch.tril(kv_mask, diagonal=0)
        kv_mask = 1 - kv_mask
        kv_mask = (
            torch.where(
                kv_mask == 1, kv_mask.new_tensor(-9223372036854775808), kv_mask
            )
            .detach()
            .long()
        )

        for i, layer in enumerate(self.layers):
            if not self.training:
                cache_k = self.cache_k[i]
                cache_v = self.cache_v[i]
                h, cache_k, cache_v = layer(
                    h, kv_mask, freqs_cis, cache_k, cache_v
                )
            else:
                h, _, _ = layer(h, kv_mask, freqs_cis)
            if not self.training:
                self.cache_k[i] = cache_k.detach()
                self.cache_v[i] = cache_v.detach()

        h = self.norm(h)
        output = self.output(h)
        return output

    @torch.no_grad()
    def generate(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        max_new_tokens: int,
        temperature: float,
        top_p: float = 1.0,
        no_repeat_ngram_size=None,
    ):
        generated_tokens = []
        for cur_pos in range(max_new_tokens):
            logits = self._forward(input_ids, attention_mask)[:, -1, :]
            if temperature > 0:
                probs = torch.softmax(logits / temperature, dim=-1)
                next_token = sample_top_p(probs, top_p)
            else:
                next_token = torch.argmax(logits, dim=-1)
            next_token = next_token.reshape(-1)
            input_ids = torch.cat([input_ids, next_token.unsqueeze(1)], dim=1)
            attention_mask = torch.cat(
                [attention_mask, torch.ones_like(next_token).unsqueeze(1)],
                dim=1,
            )
            generated_tokens.append(next_token)
        sequences = torch.concat(
            (input_ids, torch.stack(generated_tokens, dim=1)), dim=1
        )
        return sequences


def setup_model_parallel() -> Tuple[int, int]:
    local_rank = int(os.environ.get("LOCAL_RANK", -1))
    world_size = int(os.environ.get("WORLD_SIZE", -1))
    print("local_rank:", local_rank, "world_size:", world_size)

    torch.distributed.init_process_group("nccl")
    initialize_model_parallel(world_size)
    torch.cuda.set_device(local_rank)

    # seed must be the same in all processes
    torch.manual_seed(1)
    return local_rank, world_size


def setup_model_deepspeed() -> Tuple[int, int]:
    local_rank = int(os.environ.get("LOCAL_RANK", -1))
    world_size = int(os.environ.get("WORLD_SIZE", -1))

    deepspeed.init_distributed()
    torch.cuda.set_device(local_rank)

    # seed must be the same in all processes
    torch.manual_seed(1)
    return local_rank, world_size


def load_checkpoints(
    ckpt_dir: str, local_rank: int, world_size: int
) -> Tuple[dict, dict]:
    checkpoints = sorted(Path(ckpt_dir).glob("*.pth"))
    assert world_size == len(checkpoints), (
        f"Loading a checkpoint for MP={len(checkpoints)} but world "
        f"size is {world_size}"
    )
    ckpt_path = checkpoints[local_rank]
    print("Loading")
    checkpoint = torch.load(ckpt_path, map_location="cpu")
    with open(Path(ckpt_dir) / "params.json", "r") as f:
        params = json.loads(f.read())
    return checkpoint, params


def load_model(
    ckpt_dir: str,
    tokenizer_path: str,
    local_rank: int,
    world_size: int,
    froze_embeddings: bool,
    use_fairscale: bool,
    max_batch_size: int = 32,
) -> Tuple[Transformer, HFLikeTokenizer]:

    checkpoint, params = load_checkpoints(ckpt_dir, local_rank, world_size)
    model_args: ModelArgs = ModelArgs(
        max_seq_len=1024, max_batch_size=max_batch_size, **params
    )
    model_args.froze_embeddings = froze_embeddings
    model_args.use_fairscale = use_fairscale
    tokenizer = Tokenizer(model_path=tokenizer_path)
    model_args.vocab_size = tokenizer.n_words
    torch.set_default_tensor_type(torch.cuda.HalfTensor)
    model = Transformer(model_args)
    torch.set_default_tensor_type(torch.FloatTensor)
    model.load_state_dict(checkpoint, strict=False)
    tokenizer = HFLikeTokenizer(tokenizer)

    return model, tokenizer


def load_tokenizer(tokenizer_path: str):
    tokenizer = Tokenizer(model_path=tokenizer_path)
    return tokenizer


def load_tokenizer_test(tokenizer_path: Optional[str] = None):
    tokenizer = MyTokenizer(model_path=tokenizer_path)
    return tokenizer


def load_model_test(
    ckpt_dir: str,
    tokenizer_path: str,
    local_rank: int,
    world_size: int,
    froze_embeddings: bool,
    use_fairscale: bool,
    max_batch_size: int = 32,
) -> Tuple[Transformer, HFLikeTokenizer]:

    # test the model with hf tokenizer
    model_args = ModelArgs()
    model_args.froze_embeddings = froze_embeddings
    model_args.use_fairscale = use_fairscale
    tokenizer = MyTokenizer(model_path=tokenizer_path)
    model_args.vocab_size = tokenizer.n_words
    model = Transformer(model_args).cuda()
    tokenizer = HFLikeTokenizer(tokenizer)

    return model, tokenizer



import json
import os
import random
from collections import deque, namedtuple

import deepspeed
import torch
from accelerate import Accelerator
from beartype import beartype
from beartype.typing import Deque, List, Tuple, Union
from torch.utils.data import DataLoader, Dataset
from torch.optim.lr_scheduler import CosineAnnealingWarmRestarts

from chatllama.rlhf.actor import ActorModel
from chatllama.rlhf.config import (
    Config,
    ConfigActor,
    ConfigCritic,
    ConfigReward,
)
from chatllama.rlhf.model_list import hf_models
from chatllama.rlhf.model_loader import ModelLoader
from chatllama.rlhf.reward import RewardModel, CriticModel
from chatllama.rlhf.utils import TrainingStats, ConversationLog


"""
train()
┌─────────────────────────────┐
│                             │◄─────────────────────────┐
│                             │                          │
│      ┌─────────────┐        │                          │
│      │ user input  │        │                          │ learn()
│      └─────┬───────┘        │             ┌────────────┴─────────────┐
│            │                │             │                          │
│            │                │             │       ┌────────┐         │
│            │                │             │   ┌───│ Update │──┐      │
│            │                │             │   │   └────▲───┘  │      │
│   ┌────────▼────────────┐   │             │   │        │      │      │
│   │  Actor (LLM Model)  │   │             │   │     ┌──┴───┐  │      │
│   └────────┬────────────┘   │             │   │     │ PPO  │  │      │
│            │                │             │   │     └▲────▲┘  │      │
│            │                │             │   │      │    │   │      │
│            │                │             │   │      │    │   │      │
│    ┌───────▼──────┐         │             │ ┌─▼──────┴┐ ┌─┴───▼──┐   │
│    │ Reward Model │         │             │ │  Actor  │ │ Critic │   │
│    └──────────────┘         │             │ └─────────┘ └────────┘   │
│                             │             │                          │
│                             │ x Episodes  └─────────────▲────────────┘
└───────────────┬─────────────┘                           │   x Epochs
                │ store N Examples per Timestep           │  
         ┌──────▼──────┐                                  │
         │             │                                  │
         │  Memories   ├──────────────────────────────────┘
         │             │ (update timesteps x N Examples)
         └─────────────┘
"""  # noqa W291


def change_tokenization(tokens, tokenizer1, tokenizer2):
    """Change the tokenizer of the tokens

    Args:
        tokens (torch.Tensor): Tokens to be changed
        tokenizer1 (transformers.PreTrainedTokenizer): Tokenizer to be changed
        tokenizer2 (transformers.PreTrainedTokenizer): Tokenizer to be
            changed to

    Returns:
        encoded_tokens: Encoded tokens
    """

    # decode tokens
    with torch.no_grad():
        decoded_tokens = [
            tokenizer1.decode(token) for i, token in enumerate(tokens)
        ]

        # remove all the pad tokens
        decoded_tokens = [
            token.replace(tokenizer1.pad_token, "") for token in decoded_tokens
        ]

        # remove all the eos tokens
        decoded_tokens = [
            token.replace(tokenizer1.eos_token, "") for token in decoded_tokens
        ]

        # encode the actions with critic tokenizer
        encoded_tokens = tokenizer2(
            decoded_tokens,
            return_tensors="pt",
            padding=True,
            truncation=True,
        )

    return encoded_tokens


ConfigType = Union[ConfigActor, ConfigReward, ConfigCritic]


@beartype
def check_model_family(config1: ConfigType, config2: ConfigType) -> bool:
    """Check if the model family is the same for the two configs
    the model family is specified in the config.model

    Args:
        config1 (ConfigType): First config
        config2 (ConfigType): Second config

    Returns:
        bool: True if the model family is the same, False otherwise
    """

    # check if both are an hugging face models
    if (config1.model in hf_models) and (config2.model in hf_models):

        # if there is a "/" remove it from the name
        model_name1 = config1.model
        model_name2 = config2.model
        if "/" in model_name1:
            model_name1 = model_name1.split("/")[1]
        if "/" in model_name2:
            model_name2 = model_name2.split("/")[1]

        # check if the model family is the same
        return model_name1.split("-")[0] == model_name2.split("-")[0]

    # check if both are not an hugging face models
    elif (config1.model not in hf_models) and (config2.model not in hf_models):

        # for now they could be only LLaMA models
        return True
    else:
        return False


class ActorCritic(torch.nn.Module):
    """Actor Critic class stores both the actor and the critic models
    and it generates values and action for given sequences during the training
    of the actor.

    Attributes:
        actor (ActorModel): Actor model
        critic (CriticModel): Critic model
        debug (bool): enable prints for Debugging
        use_same_tokenizer (bool): if True the actor and critic use the same
            tokenizer

    Methods:
        forward: given a sequence returns action logits and values (used
            to evaluate the actor during training)
        generate: given a sequence returns action, action logits, values
            sequences and sequences masks (used to generate new sequences
            during acting phase)
    """

    def __init__(self, config: Config) -> None:
        super().__init__()
        self.config = config

        self.actor = ActorModel(config.actor)

        # check if critic must be initialized from reward model
        ModelLoader.init_critic_from_reward(config.critic)
        self.critic = CriticModel(config.critic)

        # if the actor and critic use the same tokenizer is set to True
        self.use_same_tokenizer = False

        # debug flag
        self.debug = config.actor.debug

    @beartype
    def load(self) -> None:
        """Load the model from the path.
        This method is not implemented since it relies on actor and critic
        __init__ methods to perform the loading from their respective paths
        then loaded.

        """
        pass

    @beartype
    def save(self) -> None:
        """Save the model to the path
        This method is implemented to save the actor model as result of RLHF
        in the folder actor_rl instead of actor.save() method that saves it
        in the actor folder.
        """
        # get the path to save the actor
        model_folder, model_name, path = ModelLoader.get_model_path(
            config=self.config,
            is_checkpoint=False,
        )

        # save the model
        print(f"Saving model to {path} ...")
        torch.save(
            {"state_dict": self.actor.model.state_dict()},
            path,
        )

        # get the path to save the critic model
        model_folder, model_name, path = ModelLoader.get_model_path(
            config=self.config.critic,
            is_checkpoint=False,
        )

        # save the model
        print(f"Saving model to {path} ...")
        torch.save(
            {"state_dict": self.critic.model.state_dict()},
            path,
        )

    @beartype
    def forward(
        self,
        sequences_actor: torch.Tensor,
        sequences_mask_actor: torch.Tensor,
        sequences_critic: torch.Tensor,
        sequences_mask_critic: torch.Tensor,
        action_len_actor: int,
        action_len_critic: int,
    ) -> Tuple:
        """Given the whole sequences, use the actor forward to get the logits
            for each token in the sequence and the critic forward to get the
            values for each generation step.

        Args:
            sequences_actor (torch.Tensor): Sequences composed of
                [states, actions] for the actor
            sequence_mask_actor (torch.Tensor): Mask for the sequences
                of the actor
            sequences_critic (torch.Tensor): Sequences composed of
                [states, actions] for the critic
            sequences_mask_critic (torch.Tensor): Mask for the sequences
                of the critic
            action_len_actor (int): Length of the actions in the sequences
                for the actor
            action_len_critic (int): Length of the actions in the sequences
                for the critic

        Returns:
            action_logits (torch.Tensor): Logits for the actions in the
                sequences
            values (torch.Tensor): Values for the actions in the sequences
        """

        # use a single forward on the whole sequence
        # to get pi(y | x) and ignore predicted output
        actions_logits = self.actor.forward(
            sequences_actor, sequences_mask_actor
        )

        # use the critic forward to get the values for the actions
        values = self.critic.forward(sequences_critic, sequences_mask_critic)

        # return only logits and values for the actions taken
        real_actions_logits = actions_logits[:, -action_len_actor:, :]
        real_values = values[:, -action_len_critic:]

        if self.debug:
            print("ActorCritic.forward")
            print("action_len_actor", action_len_actor)
            print("action_len_critic", action_len_critic)
            print("sequences_actor.shape", sequences_actor.shape)
            print("sequences_actor", sequences_actor)
            print("sequences_critic.shape", sequences_critic.shape)
            print("sequences_critic", sequences_critic)
            print("real_action_logits.shape", actions_logits.shape)
            print("real_action_logits", actions_logits)
            print("real_values.shape", values.shape)
            print("real_values", values)

        return (
            real_actions_logits,
            real_values,
        )

    @torch.no_grad()
    @beartype
    def generate(
        self,
        states_actor: torch.Tensor,
        states_mask_actor: torch.Tensor,
        states_critic: torch.Tensor,
    ) -> Tuple:
        """Generate actions, actions_logits, values and sequences from states

        Args:
            states_actor (torch.Tensor): States for the actor
            states_mask_actor (torch.Tensor): Mask for the states for the
                actor
            states_critic (torch.Tensor): States for the critic

        Returns:
            actions (torch.Tensor): Actions generated from the states
            actions_logits (torch.Tensor): Logits for the actions generated
                from the states (i.e. pi(y | x))
            values (torch.Tensor): Values generated by the critic model
                for the actions generated by the actor (i.e. V(x))
            sequences (torch.Tensor): Sequences generated from the states
                as [states, actions]
        """

        # generate action sequence from the actor
        actions, sequences_actor = self.actor.generate(
            states_actor, states_mask_actor
        )

        # create mask for the actor sequences
        sequences_mask_actor = (
            (sequences_actor != self.actor.tokenizer.pad_token_id)
            .to(sequences_actor.device)
            .long()
            .detach()
        )

        # get the length of the actions
        action_len_actor = actions.shape[1]

        # check if different encoding is needed for the critic
        if self.use_same_tokenizer:
            sequences_critic = sequences_actor
            sequences_mask_critic = sequences_mask_actor
            action_len_critic = action_len_actor
        else:
            encoded_critic = change_tokenization(
                sequences_actor,
                self.actor.tokenizer,
                self.critic.tokenizer,
            )
            # split the encoded_critic in tokens and maks
            sequences_critic = encoded_critic["input_ids"].to(
                sequences_actor.device,
            )
            sequences_mask_critic = (
                encoded_critic["attention_mask"]
                .to(sequences_actor.device)
                .long()
                .detach()
            )

            # compute len of actions for the critic tokenizer
            action_len_critic = states_critic.shape[1]

        # generate actions_logits and values
        actions_logits, values = self.forward(
            sequences_actor,
            sequences_mask_actor,
            sequences_critic,
            sequences_mask_critic,
            action_len_actor,
            action_len_critic,
        )
        if self.debug:
            print("ActorCritic.generate")
            print("actions shape", actions.shape)
            print("actions", actions)
            print("sequence shape", sequences_actor.shape)
            print("sequence", sequences_actor)
            print("actions_logits shape", actions_logits.shape)
            print("actions_logits", actions_logits)
            print("values shape", values.shape)
            print("values", values)

        return (
            actions,
            actions_logits,
            values,
            sequences_actor,
            sequences_mask_actor,
            sequences_critic,
            sequences_mask_critic,
            action_len_actor,
            action_len_critic,
        )


# structure to store the data for each experience
Memory = namedtuple(
    "Memory",
    [
        "states_actor",
        "actions",
        "values",
        "rewards",
        "actions_log_probs",
        "sequences_actor",
        "sequences_mask_actor",
        "sequences_critic",
        "sequences_mask_critic",
        "action_len_actor",
        "action_len_critic",
    ],
)


class ExperienceDataset(Dataset):
    """Dataset to train the actor-critic models"""

    def __init__(
        self,
        memories: Deque[Memory],
        device: torch.device,
    ) -> None:
        super().__init__()
        self.data = list(memories)
        self.device = device

    def __len__(
        self,
    ) -> int:
        return len(self.data)

    def __getitem__(self, idx) -> Tuple:
        # return the idx-th memory element as a tuple of tensors on the device
        item = (
            self.data[idx].states_actor.to(self.device),
            self.data[idx].actions.to(self.device),
            self.data[idx].values.to(self.device),
            self.data[idx].rewards.to(self.device),
            self.data[idx].actions_log_probs.to(self.device),
            self.data[idx].sequences_actor.to(self.device),
            self.data[idx].sequences_mask_actor.to(self.device),
            self.data[idx].sequences_critic.to(self.device),
            self.data[idx].sequences_mask_critic.to(self.device),
            int(self.data[idx].action_len_actor),
            int(self.data[idx].action_len_critic),
        )
        return item


class ExamplesSampler:
    """Store the prompt to be sampled to generate the examples
    read a json file with the following format:
    [
        {
            "user_input" : "",
        } ,
        ...
    ]
    Where:
        user_input: is the input of the user or directly the input of the user
            with the memory preappended (i.e. user_input + memory)
    """

    def __init__(
        self,
        path: str,
    ) -> None:
        self.path = path
        with open(path, "r") as f:
            data = json.load(f)
        self.data = [d["user_input"] for d in data]

    def sample(self, n: int) -> List:
        """Sample n examples from the data

        Args:
            n (int): Number of examples to sample
        """
        return random.sample(self.data, n)


class RLTrainer:
    """Train the actor-critic model using RL

    Attributes:
        config (Config): Configuration of the trainer
        debug (bool): Debug mode
        actorcritic (ActorCritic): Actor-critic model
        actor_optim (torch.optim): Optimizer for the actor
        critic_optim (torch.optim): Optimizer for the critic
        actor_scheduler (torch.optim.lr_scheduler): Scheduler for the actor
        critic_scheduler (torch.optim.lr_scheduler): Scheduler for the critic
        reward (RewardModel): Reward model
        training_stats (TrainingStats): Class to store training stats
        conversation_log (ConversationLog): Class to store the conversation
        examples_sampler (ExamplesSampler): Class to sample examples
        eps (float): small epsilon to avoid division by zero

    Methods:
        train: the training loop that calls the learn function after generating
            the experiences.
        learn: Learn from a batch of experiences and update the actor and the
            critic model.
        load_checkpoint: Load the checkpoint of the actor-critic model
        save_checkpoint: Save the checkpoint of the actor-critic model
    """

    def __init__(
        self,
        config: Config,
    ) -> None:

        # save config
        self.config = config

        # set debug mode
        self.debug = config.trainer.debug

        # initialize agent-critic
        self.actorcritic = ActorCritic(config)

        # initialize actor optimizer
        self.actor_optimizer = torch.optim.Adam(
            self.actorcritic.actor.parameters(), lr=config.trainer.actor_lr
        )

        # initialize critic optimizer
        self.critic_optimizer = torch.optim.Adam(
            self.actorcritic.critic.parameters(), lr=config.trainer.critic_lr
        )

        # scheduler (defined in the learn() method (i need dataset size))
        self.actor_scheduler = None
        self.critic_scheduler = None

        # initialize reward model
        self.reward = RewardModel(config.reward)

        # initialize class to store training stats
        path = ModelLoader.get_training_stats_path(config)
        self.training_stats = TrainingStats(path)
        model_folder, _, _ = ModelLoader.get_model_path(
            config,
            is_checkpoint=True,
        )
        path = os.path.join(model_folder, "conversations_log.json")
        self.conversation_log = ConversationLog(path)

        # initialize examples sampler
        self.example_sampler = ExamplesSampler(config.trainer.examples_path)

        # check if actor and critic use the same tokenizer
        self.actorcritic.use_same_tokenizer = check_model_family(
            config.actor, config.critic
        )

        # check if actor and reward use the same tokenizer
        self.use_same_tokenizer = check_model_family(
            config.actor, config.reward
        )

        # eps
        self.eps = 1e-8

    @beartype
    def save_checkpoint(
        self,
        current_episode: int,
        max_episode: int,
    ) -> None:

        print(f"Saving checkpoint for episode {current_episode+1}..")

        # get the path to save the checkpoint for the critic
        model_folder, model_name, path = ModelLoader.get_model_path(
            config=self.config.critic,
            is_checkpoint=True,
            current_epoch=current_episode,
            max_epochs=max_episode,
            max_steps=0,
        )

        # if the checkpoint already exists remove it
        if os.path.exists(path):
            os.remove(path)

        # save the checkpoint
        torch.save(
            {
                "episode": current_episode,
                "critic_state_dict": self.actorcritic.critic.state_dict(),
                "critic_optim_state_dict": self.critic_optimizer.state_dict(),
            },
            path,
        )

        # get the path to save the checkpoint for the actor
        model_folder, model_name, path = ModelLoader.get_model_path(
            config=self.config,
            is_checkpoint=True,
            current_epoch=current_episode,
            max_epochs=max_episode,
            max_steps=0,
        )

        # if the checkpoint already exists remove it
        if os.path.exists(path):
            os.remove(path)

        # save the checkpoint
        torch.save(
            {
                "episode": current_episode,
                "actor_state_dict": self.actorcritic.actor.state_dict(),
                "actor_optim_state_dict": self.actor_optimizer.state_dict(),
                "training_stats": self.training_stats,
            },
            path,
        )

    @beartype
    def load_checkpoint(
        self,
    ) -> int:

        critic_episode = -1
        actor_episode = -1

        # check if there are some checkpoint for the critic
        print("Looking for checkpoints...")
        path = ModelLoader.check_model_path(
            config=self.config.critic,
            is_checkpoint=True,
            current_epoch=None,
        )

        # if there are checkpoint
        if path is not None:

            # load the critic checkpoint
            print("Loading ...")
            try:
                checkpoint = torch.load(path)
            except Exception:
                print(
                    "Checkpoint of critic corrupted!"
                    "Try to remove the last checkpoint."
                    "Now Starting from episode 0"
                )
                return 0

            # load checkpoint into model
            critic_episode = checkpoint["episode"]
            self.actorcritic.critic.load_state_dict(
                checkpoint["critic_state_dict"]
            )
            self.critic_optimizer.load_state_dict(
                checkpoint["critic_optim_state_dict"]
            )

        # check if there are checkpoints for the actor
        print("Looking for checkpoints...")
        path = ModelLoader.check_model_path(
            config=self.config,
            is_checkpoint=True,
            current_epoch=None,
        )

        # if there are some checkpoints
        if path is not None:

            # load the actor checkpoint
            print("Loading ...")
            try:
                checkpoint = torch.load(path)
            except Exception:
                print(
                    "Checkpoint of actor corrupted!"
                    "Try to remove the last checkpoint."
                    "Now Starting from episode 0"
                )
                return 0

            # load checkpoint into the model
            actor_episode = checkpoint["episode"]
            self.actorcritic.actor.load_state_dict(
                checkpoint["actor_state_dict"]
            )
            self.actor_optimizer.load_state_dict(
                checkpoint["actor_optim_state_dict"]
            )
            self.training_stats = checkpoint["training_stats"]

        # check if there are some discrepancies between the checkpoints
        if critic_episode == actor_episode:
            # all ok start from next episode
            return critic_episode + 1
        else:
            print(
                f"There are some discrepancies between the checkpoints"
                f"of actor and critic \nactor episode: {actor_episode}"
                f"\n critic episode: {critic_episode}\n"
            )
            return min(critic_episode, actor_episode) + 1

    @beartype
    def learn(self, memories: Deque[Memory]) -> None:
        """Train the agent-critic model using RL:
        - for each batch of episodes, compute action logits and values
        - then compare action logits probs with memories one and values with
            rewards to compute the PPO loss and update the actor-critic model
        """
        print("Start to Learn...")

        # get parameters
        epochs = self.config.trainer.epochs
        actor_eps_clip = self.config.trainer.actor_eps_clip
        critic_eps_clip = self.config.trainer.critic_eps_clip
        beta_s = self.config.trainer.beta_s
        batch_size = self.config.trainer.batch_size
        device = self.config.trainer.device

        # create dataset from memories
        dataset = ExperienceDataset(memories, device)
        dataloader = DataLoader(dataset, batch_size=batch_size)

        # initialize scheduler for actor
        actor_lr = self.config.trainer.actor_lr
        self.actor_scheduler = CosineAnnealingWarmRestarts(
            self.actor_optimizer, T_0=len(dataset), eta_min=actor_lr * 0.1
        )

        # initialize scheduler for critic
        critic_lr = self.config.trainer.critic_lr
        self.critic_scheduler = CosineAnnealingWarmRestarts(
            self.critic_optimizer, T_0=len(dataset), eta_min=critic_lr * 0.1
        )

        # initialize deepspeed for actor
        if self.config.actor.deepspeed_enable:
            if self.config.actor.deepspeed_config_path is None:
                raise ValueError(
                    "DeepSpeed config path is None, but deepspeed is enabled"
                )
            if (
                os.path.exists(self.config.actor.deepspeed_config_path)
                is False
            ):
                raise ValueError(
                    f"DeepSpeed config path"
                    f"{self.config.actor.deepspeed_config_path}"
                    f"does not exist"
                )
            (
                actor_model_engine,
                self.actor_optimizer,
                self.dataloader,
                _,
            ) = deepspeed.initialize(
                args=None,
                model=self.actorcritic.actor,
                model_parameters=self.actorcritic.actor.parameters(),
                training_data=dataloader,
                config=self.config.actor.deepspeed_config_path,
            )
            self.actorcritic.actor = actor_model_engine

        # initialize deepspeed for critic
        if self.config.critic.deepspeed_enable:
            if self.config.critic.deepspeed_config_path is None:
                raise ValueError(
                    "DeepSpeed config path is None, but deepspeed is enabled"
                )
            if (
                os.path.exists(self.config.critic.deepspeed_config_path)
                is False
            ):
                raise ValueError(
                    f"DeepSpeed config path"
                    f"{self.config.critic.deepspeed_config_path}"
                    f"does not exist"
                )
            (
                critic_model_engine,
                self.critic_optimizer,
                _,
                _,
            ) = deepspeed.initialize(
                args=None,
                model=self.actorcritic.critic,
                model_parameters=self.actorcritic.critic.parameters(),
                config=self.config.critic.deepspeed_config_path,
            )
            self.actorcritic.critic = critic_model_engine

        # initialize actor accelerate
        if self.config.actor.accelerate_enable is True:
            actor_accelerator = Accelerator()
            (
                actor_model,
                self.actor_optimizer,
                self.train_dataloader,
                self.actor_scheduler,
            ) = actor_accelerator.prepare(
                self.actorcritic.actor,
                self.actor_optimizer,
                self.train_dataloader,
                self.actor_scheduler,
            )
            self.actorcritic.actor = actor_model

        # initialize critic accelerate
        if self.config.critic.accelerate_enable is True:
            critic_accelerator = Accelerator()
            (
                critic_model,
                self.critic_optimizer,
                self.critic_scheduler,
            ) = critic_accelerator.prepare(
                self.actorcritic.critic,
                self.critic_optimizer,
                self.critic_scheduler,
            )
            self.actorcritic.critic = critic_model

        # train agent-critic
        self.actorcritic.train()
        for epoch in range(epochs):
            for k, (
                states_actor,
                old_actions,
                old_values,
                rewards,
                old_actions_log_probs,
                sequences_actor,
                sequences_mask_actor,
                sequences_critic,
                sequences_mask_critic,
                action_len_actor,
                action_len_critic,
            ) in enumerate(dataloader):

                if self.debug:
                    print(
                        f"#########################################"
                        f" batch from memories {k} \n "
                        f"#########################################"
                        f"states_actor {states_actor.shape} \n"
                        f"old_actions {old_actions.shape} \n"
                        f"old_values {old_values.shape} \n"
                        f"rewards {rewards.shape} \n"
                        f"old_actions_log_probs "
                        f"{old_actions_log_probs.shape}\n"
                        f"sequences_actor {sequences_actor.shape} \n"
                        f"sequences_mask_actor "
                        f"{sequences_mask_actor.shape} \n"
                        f"sequences_critic {sequences_critic.shape} \n"
                        f"sequences_mask_critic "
                        f"{sequences_mask_critic.shape} \n"
                        f"action_len_actor {action_len_actor} \n"
                        f"action_len_critic {action_len_critic} \n"
                        f"#########################################"
                    )

                # get actor critic new probabilities and values
                actions_logits, values = self.actorcritic.forward(
                    sequences_actor,
                    sequences_mask_actor,
                    sequences_critic,
                    sequences_mask_critic,
                    action_len_actor.item(),
                    action_len_critic.item(),
                )

                # get action log prob
                actions_prob = (
                    torch.softmax(actions_logits, dim=-1).max(dim=-1).values
                )
                actions_log_prob = torch.log(actions_prob + self.eps)

                # compute entropy
                entropies = (actions_prob * actions_log_prob).sum(dim=-1)

                # compute KL divergence
                kl_div_loss = (
                    (actions_prob * (old_actions_log_probs - actions_log_prob))
                    .sum(dim=-1)
                    .mean()
                )

                # compute ratios
                ratios = (actions_log_prob - old_actions_log_probs).exp()

                # compute PPO loss
                if check_model_family(self.config.actor, self.config.critic):
                    # compute discounted rewards as in TRL
                    gamma = self.config.trainer.gamma_discounted
                    discounted_rewards = torch.zeros_like(old_values)
                    for i in range(discounted_rewards.shape[1]):
                        for j in range(i, discounted_rewards.shape[1]):
                            discounted_rewards[:, i] += (
                                gamma ** (j - i) * rewards[:, j]
                            )

                    advantages = (
                        discounted_rewards - old_values
                    )  # TRL has opposite sign for old values
                    advantages = (advantages - advantages.mean(dim=-1)) / (
                        advantages.std() + self.eps
                    )

                    surr1 = advantages * ratios
                else:
                    advantages = rewards - old_values[:, -1]
                    surr1 = advantages * ratios

                surr2 = (
                    torch.clamp(ratios, 1 - actor_eps_clip, 1 + actor_eps_clip)
                    * advantages
                )

                policy_loss = -torch.min(surr1, surr2) - beta_s * entropies
                policy_loss = policy_loss.mean()
                loss = policy_loss + kl_div_loss

                # check if loss item is NaN
                if torch.isnan(loss):
                    raise ValueError("Loss is nan")

                # update actor with loss
                if self.config.actor.deepspeed_enable:
                    actor_model_engine.backward(loss)
                    actor_model_engine.step()
                elif self.config.actor.accelerate_enable:
                    self.actor_optimizer.zero_grad()
                    actor_accelerator.backward(loss)
                    self.actor_optimizer.step()
                    self.actor_scheduler.step()
                else:
                    self.actor_optimizer.zero_grad()
                    loss.backward()
                    self.actor_optimizer.step()
                    self.actor_scheduler.step()

                # compute value loss
                # the loss is the distance between the rewards and the values
                # I want this distance to be small so that values are
                # representative of the rewards, for this reason i took the
                # maximum between the two.
                # The clip is limiting the slew-rate of values_loss_clipped
                value_loss_clipped = old_values + (values - old_values).clamp(
                    -critic_eps_clip, critic_eps_clip
                )
                value_loss1 = (value_loss_clipped - rewards) ** 2
                value_loss2 = (values - rewards) ** 2
                value_loss = torch.max(value_loss1, value_loss2).mean()

                if torch.isnan(value_loss):
                    raise ValueError("Value loss is nan")

                # upate critic
                if self.config.critic.deepspeed_enable:
                    critic_model_engine.backward(value_loss)
                    critic_model_engine.step()
                elif self.config.critic.accelerate_enable:
                    self.critic_optimizer.zero_grad()
                    critic_accelerator.backward(loss)
                    self.critic_optimizer.step()
                    self.critic_scheduler.step()
                else:
                    self.critic_optimizer.zero_grad()
                    value_loss.backward()
                    self.critic_optimizer.step()
                    self.critic_scheduler.step()

                # append the losses to the training stats
                self.training_stats.training_loss.append(
                    loss.detach().cpu().item()
                )
                self.training_stats.value_loss.append(
                    value_loss.detach().cpu().item()
                )

                # print iteration info
                print(
                    f"Epoch {epoch+1}/{epochs}",
                    f"Step {k+1}/{int(len(dataloader) / batch_size)}",
                    f"Loss {loss.detach().cpu().item():.4f}",
                    f"Value Loss {value_loss.detach().cpu().item():.4f}",
                )

        self.actorcritic.eval()
        print("End Learning")

    def train(
        self,
    ) -> None:

        print("Start RL Training")

        # initialize settings
        num_episodes = self.config.trainer.num_episodes
        max_timesteps = self.config.trainer.max_timesteps
        num_examples = self.config.trainer.num_examples
        update_timesteps = self.config.trainer.update_timesteps
        batch_size = self.config.trainer.batch_size
        checkpoint_steps = self.config.trainer.checkpoint_steps
        device = self.config.trainer.device

        # number of elements that the memories should contain when learning
        number_of_memories_per_learn_iteration = (
            num_examples * update_timesteps
        )

        # the number of memories must be a multiple of the batch size
        assert (
            number_of_memories_per_learn_iteration % batch_size == 0
        ), "The number of memories must be a multiple of the batch size"

        # the total number of timesteps done in the train() are
        total_number_of_timesteps = num_episodes * max_timesteps

        # the total timesteps done should be a multiple of the update timesteps
        assert total_number_of_timesteps % update_timesteps == 0, (
            "The number of timesteps (num_episodes*max_timesteps)"
            "must be a multiple of the update_timesteps"
        )

        # initialize memories
        memories = deque([])

        # load checkpoint
        start_episode = self.load_checkpoint()

        # if it is a new training from the start clear the conversation log
        if start_episode == 0:
            self.conversation_log.clear()

        # initialize counters
        cnt_timesteps = 0
        cnt_learn_iter = 0

        # loop over episodes and timesteps
        self.actorcritic.eval()
        for episode in range(start_episode, num_episodes):
            for timestep in range(max_timesteps):

                # print the iteration info
                print(
                    f"Episode: {episode + 1}/{num_episodes}, "
                    f"Timestep: {timestep + 1}/{max_timesteps}",
                    f"Learning Cnt: {cnt_timesteps + 1}/{update_timesteps}",
                )

                # counter used to count timesteps into memory
                cnt_timesteps += 1

                # sample num_examples examples from  example dataset
                inputs = self.example_sampler.sample(num_examples)

                # tokenize examples for the actor
                tok_inputs_act = self.actorcritic.actor.tokenizer(
                    inputs, padding=True, return_tensors="pt", truncation=True
                )

                # states are [batch_size, seq_len_of_states]
                states_actor = tok_inputs_act["input_ids"].to(device)
                states_mask_actor = tok_inputs_act["attention_mask"].to(device)

                # tokenize examples for the critic
                tok_inputs_crt = self.actorcritic.critic.tokenizer(
                    inputs, padding=True, return_tensors="pt", truncation=True
                )

                # states are [batch_size, seq_len_of_states]
                states_critic = tok_inputs_crt["input_ids"].to(device)

                # generate sequences of actions and values
                (
                    actions,
                    actions_logits,
                    values,
                    sequences_actor,
                    sequences_mask_actor,
                    sequences_critic,
                    sequences_mask_critic,
                    action_len_actor,
                    action_len_critic,
                ) = self.actorcritic.generate(
                    states_actor, states_mask_actor, states_critic
                )

                # compute action log probs
                action_prob = (
                    torch.softmax(actions_logits, dim=-1).max(dim=-1).values
                )
                actions_log_probs = torch.log(action_prob + self.eps)

                # get tokenized sequence for the reward models
                if self.use_same_tokenizer:
                    reward_sequence = sequences_actor
                    reward_mask = sequences_mask_actor
                elif check_model_family(
                    self.config.critic, self.config.reward
                ):
                    reward_sequence = sequences_critic
                    reward_mask = sequences_mask_critic
                else:
                    tokenized_responses = change_tokenization(
                        sequences_actor,
                        self.actorcritic.actor.tokenizer,
                        self.reward.tokenizer,
                    )
                    # get tokens and mask
                    reward_sequence = tokenized_responses["input_ids"].to(
                        device
                    )
                    reward_mask = tokenized_responses["attention_mask"].to(
                        device
                    )

                # compute rewards
                rewards = self.reward.forward(
                    reward_sequence,
                    reward_mask,
                )

                rewards = rewards[:, -action_len_critic:]
                reward = rewards[:, -1]

                # store memories of the episode / timestep
                for i in range(states_actor.shape[0]):
                    memories.append(
                        Memory(
                            states_actor[i, :].detach().cpu(),
                            actions[i, :].detach().cpu(),
                            values[i, :].detach().cpu(),
                            rewards[i, :].detach().cpu(),
                            actions_log_probs[i, :].detach().cpu(),
                            sequences_actor[i, :].detach().cpu(),
                            sequences_mask_actor[i, :].detach().cpu(),
                            sequences_critic[i, :].detach().cpu(),
                            sequences_mask_critic[i, :].detach().cpu(),
                            int(action_len_actor),
                            int(action_len_critic),
                        )
                    )

                # decode completions to be logged in the conversation log
                completions = [
                    self.actorcritic.actor.tokenizer.decode(action)
                    for action in actions
                ]
                # remove pad tokens from completions
                completions = [
                    c.replace(self.actorcritic.actor.tokenizer.pad_token, "")
                    for c in completions
                ]
                # remove eos tokens from completions
                completions = [
                    c.replace(self.actorcritic.actor.tokenizer.eos_token, "")
                    for c in completions
                ]
                # strange i need to force this?
                completions = [c.replace("<pad>", "") for c in completions]

                # log the memories in the conversation log
                for i in range(states_actor.shape[0]):
                    self.conversation_log.append(
                        inputs[i],
                        completions[i],
                        reward[i].detach().cpu().item(),
                        cnt_learn_iter,
                    )

                # learn from memories
                if (cnt_timesteps % update_timesteps == 0) and (
                    cnt_timesteps != 0
                ):
                    print("len memories", len(memories))
                    # self.conversation_log.show(cnt_learn_iter)
                    self.learn(memories)
                    mean_reward = sum([m.rewards[-1] for m in memories]) / len(
                        memories
                    )
                    print(f"Mean Reward: {mean_reward}")
                    memories.clear()
                    cnt_timesteps = 0
                    cnt_learn_iter += 1
                    self.conversation_log.save()

            # save checkpoints
            if (episode % checkpoint_steps == 0) and (episode != 0):
                self.save_checkpoint(
                    current_episode=episode, max_episode=num_episodes
                )
                self.conversation_log.save()

        # save the models
        self.actorcritic.save()
        print("End RL Training")



import yaml
import os
from dataclasses import dataclass

import torch
from beartype import beartype
from beartype.typing import Optional


@dataclass
class ConfigReward:
    """Config parameters for the reward model

    Attributes:
        device (torch.device): Device to be used for the reward model
        model (str): Model to be used for the reward model
        model_folder (str): Path to the folder where model are stored (used
            to load / store finetuned model or checkpoints)
        model_head_hidden_size (int): Hidden size of the reward model head
        max_sequence_length (int): Max sequence length of the reward model
        train_dataset_path (Optional[str]): Path to the training dataset.
            Default to None. To be specified only for the reward model trainig.
        validation_dataset_path (Optional[str]): Path to the validation
            dataset. Default to None. To be specified only for the reward
            model trainig.
        batch_size (Optional[int]): Batch size to train the reward model.
            Default to None. To be specified only for the reward model
            trainig.
        epochs (Optional[int]): Number of epochs to train the reward model.
            Default to None. To be specified only for the reward model
            trainig.
        iteration_per_print (Optional[int]): Number of iterations to print
            the training loss. Default to None. To be specified only for the
            reward model trainig.
        checkpoint_steps (Optional[int]): Number of steps (backProp) to
            interleave checkpoints. Default to None. To be specified only for
            the reward model trainig.
        checkpoint_name (Optional[str]): Name of the checkpoint. Default to
            None.
        lr (Optional[float]): Learning rate for the reward model. Default to
            None. To be specified only for the reward model distillation.
        llm_enable (bool): Enable reward model distillation. Default to True.
            Disable it if you dont have an API key.
        llm_model (Optional[str]): Model to be used for the reward model
            distillation. Default to "text-davinci-003".
        llm_temperature (Optional[float]): Temperature for the reward model
            distillation. Default to 0.9.
        llm_max_tokens (Optional[int]): Max tokens for the reward model
            distillation. Default to 64.
        deepspeed_enable (bool): Enable deepspeed for the reward model
            training. Default to False.
        deepspeed_config_path (str): Path to the deepspeed config file.
            Default to None.
        is_reward (bool): True if the model is a reward model. Default to True.
        accelerate_enable (bool): Enable accelerate for the reward model
        debug (bool): enable prints for Debugging
    """

    device: torch.device
    model: str
    model_folder: str
    model_head_hidden_size: int
    max_sequence_length: int
    train_dataset_path: Optional[str] = None
    validation_dataset_path: Optional[str] = None
    batch_size: Optional[int] = None
    epochs: Optional[int] = None
    iteration_per_print: Optional[int] = None
    checkpoint_steps: Optional[int] = None
    checkpoint_name: Optional[str] = None
    lr: Optional[float] = None
    llm_enable: Optional[bool] = False
    llm_model: Optional[str] = "text-davinci-003"
    llm_temperature: Optional[float] = 0.9
    llm_max_tokens: Optional[int] = 64
    deepspeed_enable: bool = False
    deepspeed_config_path: Optional[str] = None

    # critic specific parameters
    is_reward: bool = True
    accelerate_enable: bool = False

    debug: bool = False


# just for naming consistency
ConfigCritic = ConfigReward


@dataclass
class ConfigActor:
    """Config parameters for models

    Attributes:
        model (str): Model to be used for the actor
        model_folder (str): Path to the folder where model are stored (used
            to load / store finetuned model or checkpoints)
        tokenizer_path (str): Path to the folder where tokenizer are stored
        train_dataset_path (str): Path to the training dataset
        validation_dataset_path (Optional[str]): Path to the validation dataset
        froze_embeddings (bool): Froze embeddings for the actor
        use_fairscale (bool): Use fairscale module for the actor instead of
            pytorch native modules.
        max_sequence_length (int): Max sequence length for the actor
        max_tokens (int): Max tokens for actor generation
        min_tokens (int): Min tokens for actor generation
        additonal_prompt_tokens (int): Number of tokens to be used as safety
            to avoid too large sequences and to add a template to the
            dataset
        temperature (float): Temperature for the actor
        batch_size (int): Batch size to train the actor
        iteration_per_print (int): Number of iterations to print the
            training loss
        lr (float): Learning rate for the actor
        epochs (int): Number of epochs to train the actor
        checkpoint_steps (int): Number of steps (backProp) to interleave
            checkpoints.
        n_checkpoints_to_keep (int): Number of checkpoints to keep
            for the actor.
        deepspeed_enable (bool): Enable deepspeed for the actor.
            Default to False.
        deepspeed_config_path (str): Path to the deepspeed config file.
            Default to None.
        accelerate_enable (bool): Enable accelerate for the actor
        device (torch.device): Device to be used for the actor
        checkpoint_name (Optional[str]): Name of the checkpoint. Default to
            None.
        debug (bool): Enable prints for debugging
    """

    model: str
    model_folder: str
    tokenizer_path: str
    train_dataset_path: str
    validation_dataset_path: Optional[str]
    froze_embeddings: bool
    use_fairscale: bool
    max_sequence_length: int
    max_tokens: int
    min_tokens: int
    additonal_prompt_tokens: int
    temperature: float
    batch_size: int
    iteration_per_print: int
    lr: float
    epochs: int
    checkpoint_steps: int
    n_checkpoints_to_keep: int

    deepspeed_enable: bool
    deepspeed_config_path: Optional[str]

    accelerate_enable: bool

    device: torch.device
    checkpoint_name: Optional[str] = None
    debug: bool = False


@dataclass
class ConfigTrainer:
    """Config parameters for the trainer, used to configure the reinforcement
    learning training loop

    Attributes:
        actor_lr (float): Learning rate for the actor when training with
            reinforcement learning
        critic_lr (float): Learning rate for the critic when training with
            reinforcement learning
        actor_eps_clip (float): Epsilon clip for the actor
        critic_eps_clip (float): Epsilon clip for the critic
        beta_s (float): Beta for the actor and critic
        gamma (float): coefficient for the discounted rewards.
        examples_path (str): Path to the examples dataset
        num_episodes (int): Number of episodes, each episodes consist of
            a number of timesteps that are used to generate examples
            stored in the memory buffer.
        max_timesteps (int): Max timesteps for the actor and critic.
            for each timestep a set of examples are sampled and used to
            generate a completion and a reward.
        update_timesteps (int): Number of timesteps to update the actor and
            critic
        num_examples (int): Number of examples to generate for the actor
            and critic. For each iteration of timestep, num_examples are
            sampled from the prompt dataset, processed and stored in the
            memory buffer.
        batch_size (int): Batch size to train the actor and critic.
            This batch is used to aggregate the memory from the memory buffer
            for the actual training of the actor and critic models.
        epochs (int): Number of epochs to train the actor and critic.
        checkpoint_steps (int): Number of episodes to interleave checkpoints.
        device (torch.device): Device to be used for the actor and critic
        checkpoint_name (Optional[str]): Name of the checkpoint. Default to
            None.
    """

    actor_lr: int
    critic_lr: int
    actor_eps_clip: float
    critic_eps_clip: float
    beta_s: float
    gamma_discounted: float
    examples_path: str
    num_episodes: int
    max_timesteps: int
    update_timesteps: int
    num_examples: int
    batch_size: int
    epochs: int
    checkpoint_steps: int
    device: torch.device
    checkpoint_name: Optional[str] = None
    debug: bool = False


class Config:
    """Store the config parameters for the whole pipeline

    Args:
        trainer_dict (Optional[Dict]): Dictionary with the config parameters
            for the trainer. Default to None. If None, the config.yaml is
            used.
        actor_dict (Optional[Dict]): Dictionary with the config parameters
            for the actor. Default to None. If None, the config.yaml is
            used.
        critic_dict (Optional[Dict]): Dictionary with the config parameters
            for the critic. Default to None. If None, the config.yaml is
            used.
        reward_dict (Optional[Dict]): Dictionary with the config parameters
            for the reward. Default to None. If None, the config.yaml is
            used.
        device (Optional[torch.device]): Device to be used for the actor
            and critic. Default to None. If None, the device available is
            used.
        debug (Optional[bool]): Enable prints for debugging. Default to False.

    Attributes:
        trainer (ConfigTrainer): Config parameters for the trainer
        actor (ConfigActor): Config parameters for the actor
        critic (ConfigCritic): Config parameters for the critic
        reward (ConfigReward): Config parameters for the reward
    """

    @beartype
    def __init__(
        self,
        path: str,
        device: Optional[torch.device] = None,
        debug: Optional[bool] = False,
    ) -> None:

        # if not specified use the device available
        if device is None:
            if torch.cuda.is_available():
                device = torch.device("cuda")
            else:
                raise ValueError("No GPU available")
            print(f"Current device used :{str(device)}")

        if path is None or os.path.exists(path) is False:
            raise ValueError("Path to the config.yaml is not valid")

        # Read the config from yaml
        with open(path, "r") as c:
            config = yaml.safe_load(c)

        trainer_dict = config["trainer_config"]
        actor_dict = config["actor_config"]
        critic_dict = config["critic_config"]
        reward_dict = config["reward_config"]

        # Trainer Config
        trainer_dict["device"] = device
        trainer_dict["debug"] = debug
        self.trainer = ConfigTrainer(**trainer_dict)
        # Actor Config
        actor_dict["device"] = device
        actor_dict["debug"] = debug
        self.actor = ConfigActor(**actor_dict)
        # Critic Config
        critic_dict["device"] = device
        critic_dict["debug"] = debug
        self.critic = ConfigCritic(**critic_dict)
        self.critic.is_reward = False
        # Reward Config
        reward_dict["device"] = device
        reward_dict["debug"] = debug
        self.reward = ConfigReward(**reward_dict)



from pathlib import Path
from setuptools import setup, find_packages


REQUIREMENTS = [
    "accelerate",
    "beartype",
    "deepspeed",
    "einops",
    "fairscale",
    "langchain>=0.0.103",
    "torch",
    "tqdm",
    "transformers",
    "datasets",
    "openai",
    "plotly",
]

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf8")

setup(
    name="chatllama-py",
    version="0.0.3",
    packages=find_packages(),
    install_requires=REQUIREMENTS,
    long_description=long_description,
    include_package_data=True,
    long_description_content_type="text/markdown",
)



import json
import os
from beartype import beartype
from plotly import graph_objects as go


class TrainingStats:
    """Training statistics

    Attributes:
        training_loss (List): List of training losses
        training_accuracy (List): List of training accuracies
        value_loss (List): List of value losses
        validation_loss (List): List of validation losses
        validation_accuracy (List): List of validation accuracies
    """

    def __init__(self, path: str):
        """Initialize the training stats

        Args:
            path (str): Path to save the stats
        """
        self.training_loss = []
        self.training_accuracy = []
        self.value_loss = []
        self.validation_loss = []
        self.validation_accuracy = []
        self.path = path

    def plot(self):
        """Plot the training statistics using plotly"""
        fig = go.Figure()
        if len(self.training_loss) > 0:
            fig.add_trace(
                go.Scatter(y=self.training_loss, name="Training loss")
            )
        if len(self.training_accuracy) > 0:
            fig.add_trace(
                go.Scatter(y=self.training_accuracy, name="Training accuracy")
            )
        if len(self.value_loss) > 0:
            fig.add_trace(go.Scatter(y=self.value_loss, name="Value loss"))
        if len(self.validation_loss) > 0:
            fig.add_trace(
                go.Scatter(y=self.validation_loss, name="Validation loss")
            )
        if len(self.validation_accuracy) > 0:
            fig.add_trace(
                go.Scatter(
                    y=self.validation_accuracy, name="Validation accuracy"
                )
            )
        fig.update_layout(
            showlegend=True, xaxis_type="log", xaxis_title="steps"
        )
        fig.show()

    def save(
        self,
    ):
        """Save the stats"""
        if os.path.exists(self.path):
            with open(self.path, "r") as f:
                stats = json.load(f)
            stats["training_loss"].extend(self.training_loss)
            stats["training_accuracy"].extend(self.training_accuracy)
            stats["value_loss"].extend(self.value_loss)
            stats["validation_loss"].extend(self.validation_loss)
            stats["validation_accuracy"].extend(self.validation_accuracy)
        else:
            stats = {
                "training_loss": self.training_loss,
                "training_accuracy": self.training_accuracy,
                "value_loss": self.value_loss,
                "validation_loss": self.validation_loss,
                "validation_accuracy": self.validation_accuracy,
            }
        with open(self.path, "w") as f:
            json.dump(stats, f, indent=4)

    def load(
        self,
    ):
        """Load the stats"""
        with open(self.path, "r") as f:
            stats = json.load(f)
        self.training_loss = stats["training_loss"]
        self.training_accuracy = stats["training_accuracy"]
        self.value_loss = stats["value_loss"]
        self.validation_loss = stats["validation_loss"]
        self.validation_accuracy = stats["validation_accuracy"]

    def clear(
        self,
    ):
        """Clear the stats"""
        self.training_loss = []
        self.training_accuracy = []
        self.value_loss = []
        self.validation_loss = []
        self.validation_accuracy = []
        if os.path.exists(self.path):
            os.remove(self.path)


class ConversationLog:
    """Save the conversation:
    (user input, model output, rewards and learn_counter)
    during the RL training loop.
    """

    def __init__(self, path: str):
        self.conversation = []
        self.path = path
        if self.path is None:
            self.path = "./convesation_log.json"

    @beartype
    def append(
        self,
        user_input: str,
        model_output: str,
        reward: float,
        learn_counter: int,
    ):
        """Add a conversation to the log

        Args:
            user_input (str): User input / initial prompt
            model_output (str): Completion of the LLM model
            reward (float): Reward of the reward model assigned to the output
            learn_counter (int): Number of the learning iteration to
                distinguish the conversations that happens at different
                points of the training loopt
        """
        self.conversation.append(
            {
                "user_input": user_input,
                "model_output": model_output,
                "reward": reward,
                "learn_counter": learn_counter,
            }
        )

    def save(self):
        print("Saving conversations log")
        if os.path.exists(self.path):
            with open(self.path, "r") as f:
                conversation = json.load(f)
            self.conversation.extend(conversation)
        self.conversation = sorted(
            self.conversation, key=lambda x: float(x["learn_counter"])
        )
        with open(self.path, "w") as f:
            json.dump(self.conversation, f, indent=4)

    def load(self):
        with open(self.path, "r") as f:
            self.conversation = json.load(f)

    def clear(self):
        print("Clearing conversations log")
        self.conversation = []
        # remove the file in path exists
        if os.path.exists(self.path):
            os.remove(self.path)

    def show(self, current_iteration: int = None):
        """Show the conversation log

        Args:
            current_iteration (int): Current iteration of the training loop,
                if not None, print only the conversations that happened at
                <current_iteration>
        """
        for i, c in enumerate(self.conversation):
            if current_iteration is None:
                print(
                    f"##########################################\n"
                    f"Conversation {i} at learn_counter "
                    f"{c['learn_counter']}\n"
                    f"##########################################\n"
                    f"## User Input:\n\n{c['user_input']}\n\n"
                    f"## Model Output:\n\n{c['model_output']}\n\n"
                    f"## Reward: {c['reward']}\n\n"
                )
            else:
                if current_iteration == c["learn_counter"]:
                    print(
                        f"##########################################\n"
                        f"Conversation {i} at learn_counter "
                        f"{c['learn_counter']}\n"
                        f"##########################################\n"
                        f"## User Input:\n\n{c['user_input']}\n\n"
                        f"## Model Output:\n\n{c['model_output']}\n\n"
                        f"## Reward: {c['reward']}\n\n"
                    )



import os
import shutil

from beartype.typing import Union, Optional, Tuple

from chatllama.rlhf.config import (
    Config,
    ConfigActor,
    ConfigCritic,
    ConfigReward,
)
from chatllama.rlhf.model_list import hf_models

ConfigType = Union[Config, ConfigActor, ConfigCritic, ConfigReward]


class ModelLoader:
    """Class to load and save models and their checkpoints during training."""

    def __init__(
        self,
    ) -> None:
        pass

    @staticmethod
    def get_training_stats_path(config: ConfigType) -> str:
        """Method to get the path to the training stats file. Used when saving

        Args:
            config (ConfigType): the config object
        """
        model_folder, model_name, path = ModelLoader.get_model_path(
            config, is_checkpoint=True
        )
        stat_path = os.path.join(model_folder, "training_stats.json")
        return stat_path

    @staticmethod
    def look_for_last_checkpoint(
        model_folder: str,
        model_name: str,
    ) -> Optional[str]:
        """Method to look for the last checkpoint in the model folder
        checkpoint are saved as {model_name}_epoch_{current_epoch}.pt

        Args:
            model_folder (str): the folder where the checkpoints are saved
            model_name (str): the name of the model
        """
        # remove .pt to model name
        model_name = model_name.split(".")[0]
        checkpoints = [
            f for f in os.listdir(model_folder) if f.startswith(model_name)
        ]
        if len(checkpoints) == 0:
            return None
        else:
            checkpoints = sorted(checkpoints)
            # get last checkpoint
            last_checkpoint = checkpoints[-1]
            return last_checkpoint

    @staticmethod
    def look_for_checkpoint_by_name(
        model_folder: str,
        checkpoint_name: str,
    ) -> Optional[str]:
        """Method to look for a particular checkpoint in the model folder
        checkpoint are saved as
        {model_name}_epoch_{current_epoch}_steps_{current_steps}.pt

        Args:
            model_folder (str): the folder where the checkpoints are saved
            checkpoint_name (str): the name of the checkpoint
        """
        # look for a file named checkpoint_name in the model folder
        path = os.path.join(model_folder, checkpoint_name)
        if os.path.exists(path):
            return checkpoint_name
        else:
            return None

    @staticmethod
    def get_checkpoint_name(config: ConfigType) -> str:
        if isinstance(config, Config):
            return config.trainer.checkpoint_name
        else:
            return config.checkpoint_name

    @staticmethod
    def get_base_model_folder_from_config(config: ConfigType) -> str:
        if isinstance(config, ConfigActor) or isinstance(config, ConfigReward):
            return config.model_folder
        elif isinstance(config, Config):
            return config.actor.model_folder
        else:
            raise ValueError(
                "Config type not recognized during saving or loading"
            )

    @staticmethod
    def get_model_type_from_config(config: ConfigType) -> str:
        if isinstance(config, ConfigReward):
            # here use ad-hoc flag from config to distinguish between
            #  reward and critic
            if config.is_reward:
                return "reward"
            else:
                return "critic"
        elif isinstance(config, ConfigActor):
            return "actor"
        elif isinstance(config, Config):
            return "actor_rl"

    @staticmethod
    def get_model_name_from_config(config: ConfigType) -> str:
        model_name = None
        if isinstance(config, Config):
            model_name = config.actor.model
        elif isinstance(config, ConfigReward) or isinstance(
            config, ConfigActor
        ):
            model_name = config.model
        if model_name in hf_models:
            return os.path.split(model_name)[-1]
        if model_name is None:
            raise ValueError("Model name not found")
        return model_name

    @staticmethod
    def delete_old_checkpoints(
        model_folder: str, model_name: str, n_ckp_to_keep: int = 5
    ):
        """Method to discard old checkpoints, keeping only the last
        n_ckp_to_keep

        Args:
            model_folder (str): the folder where the checkpoints are saved
            model_name (str): the name of the model
            n_ckp_to_keep (int): the number of checkpoints to keep
        """

        # remove .pt to model name
        model_name = model_name.split(".")[0]
        checkpoints = [
            f for f in os.listdir(model_folder) if f.startswith(model_name)
        ]
        if len(checkpoints) == 0:
            return
        else:
            checkpoints = sorted(checkpoints)
            # check if the number of checkpoint is greater than 5
            if len(checkpoints) > n_ckp_to_keep:
                for c in checkpoints[:-n_ckp_to_keep]:
                    checkpoint_path = os.path.join(model_folder, c)
                    os.remove(checkpoint_path)

    @staticmethod
    def get_model_path(
        config: ConfigType,
        is_checkpoint: bool = False,
        current_epoch: Optional[int] = None,
        current_step: Optional[int] = None,
        max_epochs: int = 1_000_000_000,
        max_steps: int = 1_000_000_000,
    ) -> Tuple[str, str, Optional[str]]:
        """Method to get the path to the right model file. Used when saving
        the model.
        The hierarchy of the model folder is:
        -- model_folder: here store the models trained, for each type of model
                        there is a dedicated folder
            -- actor
            -- critic
            -- reward
            -- actor_rl
            -- checkpoints: here store the checkpoints during training, for
                            each type of model there is a dedicated folder
                -- actor
                -- critic
                -- reward
                -- actor_rl

        Args:
            config (ConfigType): the config object, contains info of the model
            is_checkpoint (bool): if True, the path is for a checkpoint
            current_epoch (Optional[int]): the current epoch, used to create
                the checkpoint name. If is_checkpoint is True, and
                current_epoch is None, return just the folder and the simple
                model name for the possible checkpoint.
            current_step (Optional[int]): the current step, used to create
                the checkpoint name.
            max_epochs (Optional[int]): the maximum number of epochs, used to
                create the checkpoint name.
            max_steps (Optional[int]): the maximum number of steps, used to
                create the checkpoint name.

        Returns:
            model_folder (str): the folder where the model is saved
            model_name (str): the name of the model
            path (Optional[str]): the path to the model. If is_checkpoint is
                True, and current_epoch is None, return None
        """
        model_folder = ModelLoader.get_base_model_folder_from_config(config)

        # Add the checkpoint path if necessary
        if is_checkpoint:
            model_folder = os.path.join(model_folder, "checkpoints")

        # Create the folder for the model type
        #  (Actor, Critic, Reward, Actor_RL)
        model_type = ModelLoader.get_model_type_from_config(config)
        model_folder = os.path.join(model_folder, model_type)

        # Make the path if not exists
        if os.path.exists(model_folder) is False:
            os.makedirs(model_folder)
            print(f"Model folder does not exist. Creating it: {model_folder}")

        # Create the model name
        model_name = ModelLoader.get_model_name_from_config(config)

        # If is a checkpoint and current epoch are available
        # extend the model name with the epoch, if none epoch is provided
        # just return the simple model name
        if is_checkpoint and current_epoch is not None:
            # number of characters to store the checkpoints
            n_char = max(len(str(max_epochs)), len(str(max_steps)))
            # create the string epoch such that it is always the same length
            # equalt to n_char (i.e. 00000001) necessary for sorting
            string_epoch = str(current_epoch)
            string_epoch = "0" * (n_char - len(string_epoch)) + string_epoch
            string_epoch = f"_epoch_{string_epoch}"
            if current_step is not None:
                string_step = str(current_step)
                string_step = "0" * (n_char - len(string_step)) + string_step
                string_step = f"_step_{string_step}"
                model_name = f"{model_name}{string_epoch}{string_step}.pt"
            else:
                model_name = f"{model_name}{string_epoch}.pt"
        else:
            model_name = f"{model_name}.pt"

        # if the epoch is not provided, and it is a checkpoint
        # is impossible to know the path to the file.
        # but we can know the model folder and the model name
        if is_checkpoint and current_epoch is None:
            path = None
        else:
            path = os.path.join(model_folder, model_name)
        return model_folder, model_name, path

    @staticmethod
    def check_model_path(
        config: ConfigType,
        is_checkpoint: bool = False,
        current_epoch: Optional[int] = None,
        current_step: Optional[int] = None,
    ) -> Optional[int]:
        """Method to check if the model path exists to load models
        or checkpoints.

        Args:
            config (ConfigType): the config object, contains info of the model
            is_checkpoint (bool): if True, the path is for a checkpoint
            current_epoch (Optional[int]): the current epoch.
                is is_checkpoint is True, and current_epoch is None,
                it will look for the last checkpoint and return it.

        Returns:
            path (Optional[str]): the path to the model. If is_checkpoint is
                True, and current_epoch is None, search for the last checkpoint
                and return it. If no checkpoint is found, return None.
            epoch (Optional[int]): the epoch of the checkpoint if an actual
                checkpoint is found. If no checkpoint is found, return None.
        """
        model_folder, model_name, path = ModelLoader.get_model_path(
            config,
            is_checkpoint,
            current_epoch,
        )

        # If i am looking for a checkpoint.
        if is_checkpoint and current_epoch is None:
            # If the checkpoint is specified by name use it
            checkpoint_name = ModelLoader.get_checkpoint_name(config)
            if checkpoint_name is not None:
                checkpoint = ModelLoader.look_for_checkpoint_by_name(
                    model_folder, checkpoint_name
                )
            else:
                checkpoint = ModelLoader.look_for_last_checkpoint(
                    model_folder, model_name
                )
            if checkpoint is not None:
                path = os.path.join(model_folder, checkpoint)
                # Get the epoch number from the checkpoint name

        if path is not None:
            if os.path.exists(path) is False:
                path = None

        if path is None:
            if is_checkpoint:
                checkpoint_name = ModelLoader.get_checkpoint_name(config)
                if checkpoint_name is not None:
                    print(
                        f"No checkpoint found at {model_folder} "
                        f"with name {config.checkpoint_name}"
                    )
                else:
                    print(
                        f"No previous checkpoint found at "
                        f"{model_folder} for {model_name}"
                    )
            else:
                print(
                    f"No previous model found at "
                    f"{model_folder} for model {model_name}"
                )
        else:
            if is_checkpoint:
                # the name is modelname_epoch_00000001_step_00000001.pt
                # or modelname_epoch_00000001.pt
                if "_step_" in path:
                    epoch = int(path.split("_epoch_")[-1].split("_")[0])
                    step = int(path.split("_step_")[-1].split(".")[0])
                    print(
                        f"Found checkpoint for epoch {epoch + 1},"
                        f" step {step + 1}..."
                    )
                else:
                    epoch = int(path.split("_epoch_")[-1].split(".")[0])
                    print(f"Found checkpoint for epoch {epoch + 1} ...")
            else:
                print(f"Found model at {path}")
        return path

    def init_critic_from_reward(config: ConfigCritic) -> None:
        """Method to initialize the critic from the reward model.
        If the critic folder is empty
        """

        if config.is_reward is True:
            raise ValueError(
                "The config should work for the Critic model,"
                "but the config seems to be for the Reward model"
            )

        # check that the critic folder is empty
        path = ModelLoader.check_model_path(config)
        _, _, critic_path = ModelLoader.get_model_path(config)
        if path is None:
            print("Initializing Critic from Reward model...")
            config.is_reward = True
            path = ModelLoader.check_model_path(config)
            if path is not None:
                _, _, reward_path = ModelLoader.get_model_path(config)
                # copy the file in reward_path to critic_path
                shutil.copy(reward_path, critic_path)
            else:
                print("Critic Model remains uninitialized")
        config.is_reward = False
